import vtk
from vtk.util.numpy_support import vtk_to_numpy
import os
import glob
import numpy as np
import time

import vedo
import vedo.utils as utils
import vedo.colors as colors
from vedo.assembly import Assembly
from vedo.mesh import Mesh
from vedo.pointcloud import Points
from vedo.picture import Picture
from vedo.volume import Volume
import vedo.docs as docs
import vedo.settings as settings

__doc__ = (
    """
Submodule to load/write meshes and other objects in different formats,
and other I/O functionalities.
"""
    + docs._defs
)

__all__ = [
    "load",
    "download",
    "gunzip",
    "loadStructuredPoints",
    "loadStructuredGrid",
    "loadRectilinearGrid",
    "loadUnStructuredGrid",
    "loadTransform",
    "writeTransform",
    "write",
    "exportWindow",
    "importWindow",
    "screenshot",
    "ask",
    "Video",
]



def load(inputobj, unpack=True, force=False):
    """
    Load ``Mesh``, ``Volume`` and ``Picture`` objects from file or from the web.

    The output will depend on the file extension. See examples below.
    Unzip on the fly, if it ends with `.gz`.
    Can load an object directly from a URL address.

    :param bool unpack: unpack MultiBlockData into a flat list of objects.
    :param bool force: when downloading a file ignore any previous cached downloads
        and force a new one.

    :Examples:

        .. code-block:: python

            from vedo import datadir, load, show

            # Return a Mesh object
            g = load(datadir+'250.vtk')
            show(g)

            # Return a list of 2 meshes
            g = load([datadir+'250.vtk', datadir+'270.vtk'])
            show(g)

            # Return a list of meshes by reading all files in a directory
            # (if directory contains DICOM files then a Volume is returned)
            g = load('mydicomdir/')
            show(g)

            # Return a Volume. Color/Opacity transfer functions can be specified later.
            g = load(datadir+'embryo.slc')
            g.c(['y','lb','w']).alpha((0.0, 0.4, 0.9, 1)).show()

            # Download a file from a URL address and unzip it on the fly
            g = load('https://vedo.embl.es/examples/panther.stl.gz')
            show(g)
    """
    acts = []
    if utils.isSequence(inputobj):
        flist = inputobj
    elif isinstance(inputobj, str) and inputobj.startswith('https://'):
        flist = [inputobj]
    else:
        flist = sorted(glob.glob(inputobj))

    for fod in flist:

        if fod.startswith('https://'):
            fod = download(fod, force=force, verbose=False)

        if os.path.isfile(fod): ### it's a file

            if fod.endswith('.gz'):
                fod = gunzip(fod)

            a = _load_file(fod, unpack)
            acts.append(a)

        elif os.path.isdir(fod):### it's a directory or DICOM
            flist = os.listdir(fod)
            if '.dcm' in flist[0]: ### it's DICOM
                reader = vtk.vtkDICOMImageReader()
                reader.SetDirectoryName(fod)
                reader.Update()
                image = reader.GetOutput()
                actor = Volume(image)

                actor.info['PixelSpacing'] = reader.GetPixelSpacing()
                actor.info['Width'] = reader.GetWidth()
                actor.info['Height'] = reader.GetHeight()
                actor.info['PositionPatient'] = reader.GetImagePositionPatient()
                actor.info['OrientationPatient'] = reader.GetImageOrientationPatient()
                actor.info['BitsAllocated'] = reader.GetBitsAllocated()
                actor.info['PixelRepresentation'] = reader.GetPixelRepresentation()
                actor.info['NumberOfComponents'] = reader.GetNumberOfComponents()
                actor.info['TransferSyntaxUID'] = reader.GetTransferSyntaxUID()
                actor.info['RescaleSlope'] = reader.GetRescaleSlope()
                actor.info['RescaleOffset'] = reader.GetRescaleOffset()
                actor.info['PatientName'] = reader.GetPatientName()
                actor.info['StudyUID'] = reader.GetStudyUID()
                actor.info['StudyID'] = reader.GetStudyID()
                actor.info['GantryAngle'] = reader.GetGantryAngle()

                acts.append(actor)

            else: ### it's a normal directory
                utils.humansort(flist)
                for ifile in flist:
                    a = _load_file(fod+'/'+ifile, unpack)
                    acts.append(a)
        else:
            colors.printc("\times Error in load(): cannot find", fod, c='r')

    if len(acts) == 1:
        if not acts[0]:
            colors.printc("\times Error in load(): cannot load", inputobj, c='r')
        # settings.collectable_actors.append(acts[0])
        return acts[0]
    elif len(acts) == 0:
        colors.printc("\times Error in load(): cannot load", inputobj, c='r')
        return None
    else:
        # settings.collectable_actors += acts
        return acts


def _load_file(filename, unpack):
    fl = filename.lower()

    ################################################################# other formats:
    if fl.endswith(".xml") or fl.endswith(".xml.gz") or fl.endswith(".xdmf"):
        # Fenics tetrahedral file
        actor = loadDolfin(filename)
    elif fl.endswith(".neutral") or fl.endswith(".neu"):  # neutral tetrahedral file
        actor = loadNeutral(filename)
    elif fl.endswith(".gmsh"):  # gmesh file
        actor = loadGmesh(filename)
    elif fl.endswith(".pcd"):  # PCL point-cloud format
        actor = loadPCD(filename)
        actor.GetProperty().SetPointSize(2)
    elif fl.endswith(".off"):
        actor = loadOFF(filename)
    elif fl.endswith(".3ds"):  # 3ds format
        actor = load3DS(filename)
    elif fl.endswith(".wrl"):
        importer = vtk.vtkVRMLImporter()
        importer.SetFileName(filename)
        importer.Read()
        importer.Update()
        actors = importer.GetRenderer().GetActors() #vtkActorCollection
        actors.InitTraversal()
        wacts = []
        for i in range(actors.GetNumberOfItems()):
            act = actors.GetNextActor()
            wacts.append(act)
        actor = Assembly(wacts)

        ################################################################# volumetric:
    elif fl.endswith(".tif") or fl.endswith(".slc") or fl.endswith(".vti") \
        or fl.endswith(".mhd") or fl.endswith(".nrrd") or fl.endswith(".nii") \
        or fl.endswith(".dem"):
        img = loadImageData(filename)
        actor = Volume(img)

        ################################################################# 2D images:
    elif fl.endswith(".png") or fl.endswith(".jpg") \
        or fl.endswith(".bmp") or fl.endswith(".jpeg") or fl.endswith(".gif"):
        if ".png" in fl:
            picr = vtk.vtkPNGReader()
        elif ".jpg" in fl or ".jpeg" in fl:
            picr = vtk.vtkJPEGReader()
        elif ".bmp" in fl:
            picr = vtk.vtkBMPReader()
        elif ".gif" in fl:
            from PIL import Image, ImageSequence
            img = Image.open(filename)
            frames = []
            for frame in ImageSequence.Iterator(img):
                a = np.array(frame.convert('RGB').getdata(), dtype=np.uint8)
                a = a.reshape(frame.size[1], frame.size[0],3)
                frames.append(Picture(a))
            return frames

        picr.SetFileName(filename)
        picr.Update()
        actor = Picture(picr.GetOutput())  # object derived from vtk.vtkImageActor()

        ################################################################# multiblock:
    elif fl.endswith(".vtm") or fl.endswith(".vtmb"):
        read = vtk.vtkXMLMultiBlockDataReader()
        read.SetFileName(filename)
        read.Update()
        mb = read.GetOutput()
        if unpack:
            acts = []
            for i in range(mb.GetNumberOfBlocks()):
                b =  mb.GetBlock(i)
                if isinstance(b, (vtk.vtkPolyData,
                                  vtk.vtkUnstructuredGrid,
                                  vtk.vtkStructuredGrid,
                                  vtk.vtkRectilinearGrid)):
                    acts.append(Mesh(b))
                elif isinstance(b, vtk.vtkImageData):
                    acts.append(Volume(b))
                elif isinstance(b, vtk.vtkUnstructuredGrid):
                    acts.append(vedo.UGrid(b))
            return acts
        else:
            return mb

        ################################################################# numpy:
    elif fl.endswith(".npy") or fl.endswith(".npz"):
        acts = loadNumpy(filename)

        if unpack is False:
            return Assembly(acts)
        return acts

    elif fl.endswith(".geojson"):
        return loadGeoJSON(filename)

    elif fl.endswith(".pvd"):
        return loadPVD(filename)

    elif fl.endswith(".pdb"):
        return loadPDB(filename)

        ################################################################# polygonal mesh:
    else:
        if fl.endswith(".vtk"): # read all legacy vtk types

            #output can be:
            # PolyData, StructuredGrid, StructuredPoints, UnstructuredGrid, RectilinearGrid
            reader = vtk.vtkDataSetReader()
            reader.ReadAllScalarsOn()
            reader.ReadAllVectorsOn()
            reader.ReadAllTensorsOn()
            reader.ReadAllFieldsOn()
            reader.ReadAllNormalsOn()
            reader.ReadAllColorScalarsOn()

        elif fl.endswith(".ply"):
            reader = vtk.vtkPLYReader()
        elif fl.endswith(".obj"):
            reader = vtk.vtkOBJReader()
        elif fl.endswith(".stl"):
            reader = vtk.vtkSTLReader()
        elif fl.endswith(".byu") or fl.endswith(".g"):
            reader = vtk.vtkBYUReader()
        elif fl.endswith(".foam"):  # OpenFoam
            reader = vtk.vtkOpenFOAMReader()
        elif fl.endswith(".pvd"):
            reader = vtk.vtkXMLGenericDataObjectReader()
        elif fl.endswith(".vtp"):
            reader = vtk.vtkXMLPolyDataReader()
        elif fl.endswith(".vts"):
            reader = vtk.vtkXMLStructuredGridReader()
        elif fl.endswith(".vtu"):
            reader = vtk.vtkXMLUnstructuredGridReader()
        elif fl.endswith(".vtr"):
            reader = vtk.vtkXMLRectilinearGridReader()
        elif fl.endswith(".pvtk"):
            reader = vtk.vtkPDataSetReader()
        elif fl.endswith(".pvtr"):
            reader = vtk.vtkXMLPRectilinearGridReader()
        elif fl.endswith("pvtu"):
            reader = vtk.vtkXMLPUnstructuredGridReader()
        elif fl.endswith(".txt") or fl.endswith(".xyz"):
            reader = vtk.vtkParticleReader()  # (format is x, y, z, scalar)
        elif fl.endswith(".facet"):
            reader = vtk.vtkFacetReader()
        else:
            return None

        reader.SetFileName(filename)
        reader.Update()
        routput = reader.GetOutput()

        if not routput:
            colors.printc("\noentry Unable to load", filename, c='r')
            return None

        if isinstance(routput, vtk.vtkUnstructuredGrid):
            actor = vedo.TetMesh(routput)

        else:
            actor = Mesh(routput)
            if fl.endswith(".txt") or fl.endswith(".xyz"):
                actor.GetProperty().SetPointSize(4)

    actor.filename = filename
    actor.fileSize, actor.created = fileInfo(filename)
    return actor


def download(url, force=False, verbose=True):
    """Retrieve a file from a url, save it locally and return its path."""

    if not url.startswith('https://'):
        colors.printc('Invalid URL (must start with https):\n', url, c='r')
        return url
    url = url.replace('www.dropbox', 'dl.dropbox')

    basename = os.path.basename(url)

    if '?' in basename:
        basename = basename.split('?')[0]

    from tempfile import NamedTemporaryFile
    tmp_file = NamedTemporaryFile(delete=False)
    tmp_file.name = os.path.join(os.path.dirname(tmp_file.name),
                                 os.path.basename(basename))

    if force==False and os.path.exists(tmp_file.name):
        if verbose:
            colors.printc("using cached file:", tmp_file.name)
            #colors.printc("     (use force=True to force a new download)")
        return tmp_file.name

    try:
        from urllib.request import urlopen
    except ImportError:
        import urllib2
        import contextlib
        urlopen = lambda url_: contextlib.closing(urllib2.urlopen(url_))

    if verbose:
        colors.printc('reading', basename, 'from',
                      url.split('/')[2][:40],'...', end='')

    with urlopen(url) as response, open(tmp_file.name, 'wb') as output:
        output.write(response.read())

    if verbose: colors.printc(' done.')
    return tmp_file.name


def gunzip(filename):
    """Unzip a ``.gz`` file to a temporary file and returns its path."""
    if not filename.endswith('.gz'):
        #colors.printc("gunzip() error: file must end with .gz", c='r')
        return filename
    from tempfile import NamedTemporaryFile
    import gzip

    tmp_file = NamedTemporaryFile(delete=False)
    tmp_file.name = os.path.join(os.path.dirname(tmp_file.name),
                                 os.path.basename(filename).replace('.gz',''))
    inF = gzip.open(filename, "rb")
    outF = open(tmp_file.name, "wb")
    outF.write(inF.read())
    outF.close()
    inF.close()
    return tmp_file.name


def fileInfo(file_path):
    sz, created= "", ""
    if os.path.isfile(file_path):
        file_info = os.stat(file_path)
        num = file_info.st_size
        for x in ['B', 'KB', 'MB', 'GB', 'TB']:
            if num < 1024.0:
                break
            num /= 1024.0
        sz =  "%3.1f%s" % (num, x)
        created = time.ctime(os.path.getmtime(file_path))
    return sz, created


###################################################################
def loadStructuredPoints(filename):
    """Load and return a ``vtkStructuredPoints`` object from file."""
    reader = vtk.vtkStructuredPointsReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadStructuredGrid(filename):
    """Load and return a ``vtkStructuredGrid`` object from file."""
    if filename.endswith(".vts"):
        reader = vtk.vtkXMLStructuredGridReader()
    else:
        reader = vtk.vtkStructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadUnStructuredGrid(filename):
    """Load and return a ``vtkunStructuredGrid`` object from file."""
    if filename.endswith(".vtu"):
        reader = vtk.vtkXMLUnstructuredGridReader()
    else:
        reader = vtk.vtkUnstructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadRectilinearGrid(filename):
    """Load and return a ``vtkRectilinearGrid`` object from file."""
    if filename.endswith(".vtr"):
        reader = vtk.vtkXMLRectilinearGridReader()
    else:
        reader = vtk.vtkRectilinearGridReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadXMLData(filename):
    """Read any type of vtk data object encoded in XML format."""
    reader = vtk.vtkXMLGenericDataObjectReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


###################################################################
def load3DS(filename):
    """Load ``3DS`` file format from file. Return an ``Assembly(vtkAssembly)`` object."""
    renderer = vtk.vtkRenderer()
    renWin = vtk.vtkRenderWindow()
    renWin.AddRenderer(renderer)

    importer = vtk.vtk3DSImporter()
    importer.SetFileName(filename)
    importer.ComputeNormalsOn()
    importer.SetRenderWindow(renWin)
    importer.Update()

    actors = renderer.GetActors()  # vtkActorCollection
    acts = []
    for i in range(actors.GetNumberOfItems()):
        a = actors.GetItemAsObject(i)
        acts.append(a)
    del renWin
    return Assembly(acts)


def loadOFF(filename):
    """Read the OFF file format."""
    f = open(filename, "r")
    lines = f.readlines()
    f.close()

    vertices = []
    faces = []
    NumberOfVertices = None
    i = -1
    for text in lines:
        if len(text) == 0:
            continue
        if text == '\n':
            continue
        if "#" in text:
            continue
        if "OFF" in text:
            continue

        ts = text.split()
        n = len(ts)

        if not NumberOfVertices and n > 1:
            NumberOfVertices, NumberOfFaces = int(ts[0]), int(ts[1])
            continue
        i += 1

        if i < NumberOfVertices and n == 3:
            x, y, z = float(ts[0]), float(ts[1]), float(ts[2])
            vertices.append([x, y, z])

        ids = []
        if NumberOfVertices <= i < (NumberOfVertices + NumberOfFaces + 1) and n > 2:
            ids += [int(xx) for xx in ts[1:]]
            faces.append(ids)

    return Mesh(utils.buildPolyData(vertices, faces))


def loadGeoJSON(filename):
    """Load GeoJSON files."""
    jr = vtk.vtkGeoJSONReader()
    jr.SetFileName(filename)
    jr.Update()
    return Mesh(jr.GetOutput())


def loadDolfin(filename, exterior=False):
    """Reads a `Fenics/Dolfin` file format (.xml or .xdmf).
    Return an ``Mesh`` object."""
    import sys
    if sys.version_info[0] < 3:
        return _loadDolfin_old(filename)

    import dolfin

    if filename.lower().endswith('.xdmf'):
        f = dolfin.XDMFFile(filename)
        m = dolfin.Mesh()
        f.read(m)
    else:
        m = dolfin.Mesh(filename)

    bm = dolfin.BoundaryMesh(m, "exterior")

    if exterior:
        poly = utils.buildPolyData(bm.coordinates(), bm.cells(), fast=True, tetras=True)
    else:
        polyb = utils.buildPolyData(bm.coordinates(), bm.cells(), fast=True, tetras=True)
        polym = utils.buildPolyData(m.coordinates(), m.cells(), fast=True, tetras=True)
        app = vtk.vtkAppendPolyData()
        app.AddInputData(polym)
        app.AddInputData(polyb)
        app.Update()
        poly = app.GetOutput()
    return Mesh(poly).lw(0.1)


def _loadDolfin_old(filename, exterior='dummy'):
    import xml.etree.ElementTree as et

    if filename.endswith(".gz"):
        import gzip

        inF = gzip.open(filename, "rb")
        outF = open("/tmp/filename.xml", "wb")
        outF.write(inF.read())
        outF.close()
        inF.close()
        tree = et.parse("/tmp/filename.xml")
    else:
        tree = et.parse(filename)

    coords, faces = [], []
    for mesh in tree.getroot():
        for elem in mesh:
            for e in elem.findall("vertex"):
                x = float(e.get("x"))
                y = float(e.get("y"))
                ez = e.get("z")
                if ez is None:
                    coords.append([x, y])
                else:
                    z = float(ez)
                    coords.append([x, y, z])

            tets = elem.findall("tetrahedron")
            if not len(tets):
                tris = elem.findall("triangle")
                for e in tris:
                    v0 = int(e.get("v0"))
                    v1 = int(e.get("v1"))
                    v2 = int(e.get("v2"))
                    faces.append([v0, v1, v2])
            else:
                for e in tets:
                    v0 = int(e.get("v0"))
                    v1 = int(e.get("v1"))
                    v2 = int(e.get("v2"))
                    v3 = int(e.get("v3"))
                    faces.append([v0, v1, v2, v3])

    poly = utils.buildPolyData(coords, faces)
    return Mesh(poly)


def loadPVD(filename):
    """Reads a paraview set of files."""
    import xml.etree.ElementTree as et

    tree = et.parse(filename)

    dname = os.path.dirname(filename)
    if not dname:
        dname = '.'

    listofobjs = []
    for coll in tree.getroot():
        for dataset in coll:
            fname = dataset.get("file")
            ob = load(dname+'/'+fname)
            tm = dataset.get("timestep")
            if tm:
                ob.time(tm)
            listofobjs.append(ob)
    if len(listofobjs) == 1:
        return listofobjs[0]
    elif len(listofobjs) == 0:
        return None
    else:
        return listofobjs


def loadPDB(filename, bondScale=1, hydrogenBondScale=1, coilWidth=0.3, helixWidth=1.3):
    """Reads a molecule Protein Data Bank file."""
    rr = vtk.vtkPDBReader()
    rr.SetFileName('1btn.pdb')
    rr.SetBScale(bondScale)
    rr.SetHBScale(hydrogenBondScale)
    rr.Update()
    prf = vtk.vtkProteinRibbonFilter()
    prf.SetCoilWidth(coilWidth)
    prf.SetHelixWidth(helixWidth)
    prf.SetInputData(rr.GetOutput())
    prf.Update()
    return Mesh(prf.GetOutput())


def loadNeutral(filename):
    """Reads a `Neutral` tetrahedral file format. Return an ``Mesh`` object."""
    f = open(filename, "r")
    lines = f.readlines()
    f.close()

    ncoords = int(lines[0])
    coords = []
    for i in range(1, ncoords + 1):
        x, y, z = lines[i].split()
        coords.append([float(x), float(y), float(z)])

    ntets = int(lines[ncoords + 1])
    idolf_tets = []
    for i in range(ncoords + 2, ncoords + ntets + 2):
        text = lines[i].split()
        v0, v1, v2, v3 = int(text[1])-1, int(text[2])-1, int(text[3])-1, int(text[4])-1
#        p0, p1, p2, p3 = np.array(coords[v1]), np.array(coords[v0]), coords[v3], coords[v2]
#        d10 = p1-p0
#        d21 = p2-p1
#        dc = np.cross(d10, d21)
#        print(np.dot(dc,p3-p0))
        idolf_tets.append([v0, v1, v2, v3])

    poly = utils.buildPolyData(coords, idolf_tets)
    return Mesh(poly)


def loadGmesh(filename):
    """Reads a `gmesh` file format. Return an ``Mesh`` object."""
    f = open(filename, "r")
    lines = f.readlines()
    f.close()

    nnodes = 0
    index_nodes = 0
    for i, line in enumerate(lines):
        if "$Nodes" in line:
            index_nodes = i + 1
            nnodes = int(lines[index_nodes])
            break
    node_coords = []
    for i in range(index_nodes + 1, index_nodes + 1 + nnodes):
        cn = lines[i].split()
        node_coords.append([float(cn[1]), float(cn[2]), float(cn[3])])

    nelements = 0
    index_elements = 0
    for i, line in enumerate(lines):
        if "$Elements" in line:
            index_elements = i + 1
            nelements = int(lines[index_elements])
            break
    elements = []
    for i in range(index_elements + 1, index_elements + 1 + nelements):
        ele = lines[i].split()
        elements.append([int(ele[-3]), int(ele[-2]), int(ele[-1])])

    poly = utils.buildPolyData(node_coords, elements, indexOffset=1)
    return Mesh(poly)


def loadPCD(filename):
    """Return a ``Mesh`` made of only vertex points
    from `Point Cloud` file format. Return an ``Mesh`` object."""
    f = open(filename, "r")
    lines = f.readlines()
    f.close()
    start = False
    pts = []
    N, expN = 0, 0
    for text in lines:
        if start:
            if N >= expN:
                break
            l = text.split()
            pts.append([float(l[0]), float(l[1]), float(l[2])])
            N += 1
        if not start and "POINTS" in text:
            expN = int(text.split()[1])
        if not start and "DATA ascii" in text:
            start = True
    if expN != N:
        colors.printc("Mismatch in pcd file", expN, len(pts), c="red")
    poly = utils.buildPolyData(pts)
    return Mesh(poly).pointSize(4)


def toNumpy(obj):
    '''Dump a vedo object to numpy format.'''

    adict = dict()
    adict['type'] = 'unknown'

    ########################################################
    def _fillcommon(obj, adict):
        adict['filename'] = obj.filename
        adict['name'] = obj.name
        adict['time'] = obj.time()
        adict['rendered_at'] = obj.renderedAt
        adict['position'] = obj.pos()
        adict['info'] = obj.info
        m = np.eye(4)
        vm = obj.getTransform().GetMatrix()
        for i in [0, 1, 2, 3]:
            for j in [0, 1, 2, 3]:
                m[i,j] = vm.GetElement(i, j)
        adict['transform'] = m
        minv = np.eye(4)
        vm.Invert()
        for i in [0, 1, 2, 3]:
            for j in [0, 1, 2, 3]:
                minv[i,j] = vm.GetElement(i, j)
        adict['transform_inverse'] = minv

    ########################################################
    def _fillmesh(obj, adict):

        adict['points'] = obj.points(transformed=False).astype(np.float32)
        poly = obj.polydata()
        adict['flagText'] = obj.flagText

        adict['cells'] = None
        if poly.GetNumberOfPolys():
            try:
                adict['cells'] = np.array(obj.faces(), dtype=np.uint32)
            except ValueError:
                adict['cells'] = obj.faces()

        adict['lines'] = None
        if poly.GetNumberOfLines():
            adict['lines'] = obj.lines()

        adict['pointdata'] = []
        for iname in obj.getArrayNames()['PointData']:
            if 'Normals' in iname.lower(): continue
            arr = poly.GetPointData().GetArray(iname)
            adict['pointdata'].append([vtk_to_numpy(arr), iname])
        adict['celldata'] = []
        for iname in obj.getArrayNames()['CellData']:
            if 'Normals' in iname.lower(): continue
            arr = poly.GetCellData().GetArray(iname)
            adict['celldata'].append([vtk_to_numpy(arr), iname])

        adict['activedata'] = None
        if poly.GetPointData().GetScalars():
            adict['activedata'] = ['pointdata', poly.GetPointData().GetScalars().GetName()]
        elif poly.GetCellData().GetScalars():
            adict['activedata'] = ['celldata',  poly.GetCellData().GetScalars().GetName()]

        adict['LUT'] = None
        adict['LUT_range'] = None
        lut = obj._mapper.GetLookupTable()
        if lut:
            nlut = lut.GetNumberOfTableValues()
            lutvals=[]
            for i in range(nlut):
                v4 = lut.GetTableValue(i) # r, g, b, alpha
                lutvals.append(v4)
            adict['LUT'] = lutvals
            adict['LUT_range'] = lut.GetRange()

        prp = obj.GetProperty()
        adict['alpha'] = prp.GetOpacity()
        adict['representation'] = prp.GetRepresentation()
        adict['pointsize'] = prp.GetPointSize()

        adict['linecolor'] = None
        adict['linewidth'] = None
        if prp.GetEdgeVisibility():
            adict['linewidth'] = obj.lineWidth()
            adict['linecolor'] = obj.lineColor()

        adict['ambient'] = prp.GetAmbient()
        adict['diffuse'] = prp.GetDiffuse()
        adict['specular'] = prp.GetSpecular()
        adict['specularpower'] = prp.GetSpecularPower()
        adict['specularcolor'] = prp.GetSpecularColor()
        adict['shading'] = prp.GetInterpolation()
        adict['color'] = prp.GetColor()
        adict['lightingIsOn'] = prp.GetLighting()
        adict['backColor'] = None
        if obj.GetBackfaceProperty():
            adict['backColor'] = obj.GetBackfaceProperty().GetColor()

        adict['scalarvisibility'] = obj.mapper().GetScalarVisibility()

        adict['texture'] = None

    ######################################################## Mesh
    if isinstance(obj, Points):
        adict['type'] = 'Mesh'
        _fillcommon(obj, adict)
        _fillmesh(obj, adict)

    ######################################################## Assembly
    elif isinstance(obj, Assembly):
        pass
        # adict['type'] = 'Assembly'
        # _fillcommon(obj, adict)
        # adict['actors'] = []
        # for a in obj.unpack():
        #     assdict = dict()
        #     if isinstance(a, Mesh):
        #         _fillmesh(a, assdict)
        #         adict['actors'].append(assdict)

    ######################################################## Volume
    elif isinstance(obj, Volume):
        adict['type'] = 'Volume'
        _fillcommon(obj, adict)
        imgdata = obj.inputdata()
        arr = vtk_to_numpy(imgdata.GetPointData().GetScalars())
        adict['array'] = arr.reshape(imgdata.GetDimensions())
        adict['mode'] = obj.mode()
        #adict['jittering'] = obj.mapper().GetUseJittering()

        prp = obj.GetProperty()
        ctf = prp.GetRGBTransferFunction()
        otf = prp.GetScalarOpacity()
        gotf = prp.GetGradientOpacity()
        smin, smax = ctf.GetRange()
        xs = np.linspace(smin, smax, num=100, endpoint=True)
        cols, als, algrs = [], [], []
        for x in xs:
            cols.append(ctf.GetColor(x))
            als.append(otf.GetValue(x))
            if gotf:
                algrs.append(gotf.GetValue(x))
        adict['color'] = cols
        adict['alpha'] = als
        adict['alphagrad'] = algrs

    ######################################################## Picture
    elif isinstance(obj, Picture):
        adict['type'] = 'Picture'
        _fillcommon(obj, adict)
        adict['array'] = vtk_to_numpy(obj.inputdata().GetPointData().GetScalars())
        adict['shape'] = obj.inputdata().GetDimensions()
        print('toNumpy(): vedo.Picture', obj.shape, obj.GetPosition())

    ######################################################## Text2D
    elif isinstance(obj, vedo.Text2D):
        adict['type'] = 'Text2D'
        adict['rendered_at'] = obj.renderedAt
        adict['text'] = obj.text()
        adict['position'] = obj.GetPosition()
        adict['color'] = obj.property.GetColor()
        adict['font'] =  obj.font()
        adict['size']  = obj.property.GetFontSize()/22.5
        adict['bgcol'] = obj.property.GetBackgroundColor()
        adict['alpha'] = obj.property.GetBackgroundOpacity()
        adict['frame'] = obj.property.GetFrame()
        # print('toNumpy(): vedo.Text2D', obj.text()[:10], obj.font(), obj.GetPosition())

    else:
        pass
        #colors.printc('Unknown object type in toNumpy()', [obj], c='r')

    return adict


def loadNumpy(inobj):
    """Load a vedo format file or scene."""

    # make sure the numpy file is not containing a scene
    if isinstance(inobj, str): # user passing a file

        if inobj.endswith('.npy'):
            data = np.load(inobj, allow_pickle=True, encoding='latin1').flatten()
        elif  inobj.endswith('.npz'):
            data = np.load(inobj, allow_pickle=True)['vedo_scenes']
            #print(data.files)

        if "objects" in data[0].keys():  # loading a full scene!!
            return importWindow(data[0]) # <-------

        if 'type' not in data[0].keys():
            return data[0] # it's a very normal numpy data object, just return it!

    else:
        data = inobj


    ######################################################
    def _loadcommon(obj, d):
        keys = d.keys()
        if 'time' in keys: obj.time(d['time'])
        if 'name' in keys: obj.name = d['name']
        if 'filename' in keys: obj.filename = d['filename']

        if 'transform' in keys and len(d['transform']) == 4:
            vm = vtk.vtkMatrix4x4()
            for i in [0, 1, 2, 3]:
               for j in [0, 1, 2, 3]:
                   vm.SetElement(i, j, d['transform'][i,j])
            obj.applyTransform(vm)
        elif 'position' in keys:
            obj.pos(d['position'])

    ######################################################
    def _buildmesh(d):
        keys = d.keys()

        vertices = d['points']
        if not len(vertices):
            return None

        cells = None
        if 'cells' in keys:
            cells = d['cells']

        lines = None
        if 'lines' in keys:
            lines = d['lines']

        poly = utils.buildPolyData(vertices, cells, lines)
        msh = Mesh(poly)
        _loadcommon(msh, d)

        prp = msh.GetProperty()
        if 'ambient' in keys:        prp.SetAmbient(d['ambient'])
        if 'diffuse' in keys:        prp.SetDiffuse(d['diffuse'])
        if 'specular' in keys:       prp.SetSpecular(d['specular'])
        if 'specularpower' in keys:  prp.SetSpecularPower(d['specularpower'])
        if 'specularcolor' in keys:  prp.SetSpecularColor(d['specularcolor'])
        if 'lightingIsOn' in keys:   prp.SetLighting(d['lightingIsOn'])
        if 'shading' in keys:        prp.SetInterpolation(d['shading'])
        if 'alpha' in keys:          prp.SetOpacity(d['alpha'])
        if 'opacity' in keys:        prp.SetOpacity(d['opacity']) # synonym
        if 'representation' in keys: prp.SetRepresentation(d['representation'])
        if 'pointsize' in keys and d['pointsize']: prp.SetPointSize(d['pointsize'])

        if 'linewidth' in keys and d['linewidth']: msh.lineWidth(d['linewidth'])
        if 'linecolor' in keys and d['linecolor']: msh.lineColor(d['linecolor'])

        if 'color' in keys and d['color'] is not None:
            msh.color(d['color'])
        if 'backColor' in keys and d['backColor'] is not None:
            msh.backColor(d['backColor'])

        if 'flagText' in keys and d['flagText']:   msh.flag(d['flagText'])

        if 'celldata' in keys:
            for csc, cscname in d['celldata']:
                msh.addCellArray(csc, cscname)
        if 'pointdata' in keys:
            for psc, pscname in d['pointdata']:
                msh.addPointArray(psc, pscname)
        msh.mapper().ScalarVisibilityOff()     # deactivate scalars

        if 'LUT' in keys and 'activedata' in keys and d['activedata']:
            # print(d['activedata'],'', msh.filename)
            lut_list = d['LUT']
            ncols = len(lut_list)
            lut = vtk.vtkLookupTable()
            lut.SetNumberOfTableValues(ncols)
            lut.SetRange(d['LUT_range'])
            for i in range(ncols):
                r, g, b, a = lut_list[i]
                lut.SetTableValue(i, r, g, b, a)
            lut.Build()
            msh.mapper().SetLookupTable(lut)
            msh.mapper().ScalarVisibilityOn()  # activate scalars
            msh.mapper().SetScalarRange(d['LUT_range'])
            if d['activedata'][0] == 'celldata':
                poly.GetCellData().SetActiveScalars(d['activedata'][1])
            if d['activedata'][0] == 'pointdata':
                poly.GetPointData().SetActiveScalars(d['activedata'][1])

        if 'shading' in keys and int(d['shading'])>0:
            msh.computeNormals(cells=0) # otherwise cannot renderer phong

        if 'scalarvisibility' in keys:
            if d['scalarvisibility']:
                msh.mapper().ScalarVisibilityOn()
            else:
                msh.mapper().ScalarVisibilityOff()

        if 'texture' in keys and d['texture']:
            msh.texture(d['texture'])

        return msh
    ######################################################

    objs = []
    for d in data:
        #print('loadNumpy:', d)

        ### Mesh
        if 'mesh' == d['type'].lower():
            a = _buildmesh(d)
            if a: objs.append(a)

        ### Assembly
        elif 'assembly' == d['type'].lower():
            assacts = []
            for ad in d['actors']:
                assacts.append(_buildmesh(ad))
            asse = Assembly(assacts)
            _loadcommon(asse, d)
            objs.append(asse)

        ### Volume
        elif 'volume' == d['type'].lower():
            vol = Volume(d['array'])
            _loadcommon(vol, d)
            if 'jittering' in d.keys(): vol.jittering(d['jittering'])
            #print(d['mode'])
            vol.mode(d['mode'])
            vol.color(d['color'])
            vol.alpha(d['alpha'])
            vol.alphaGradient(d['alphagrad'])
            objs.append(vol)

        ### Picture
        elif 'picture' == d['type'].lower():
            shp = d['shape'][1], d['shape'][0]
            arr0 = d['array']
            rcv = arr0[:,0].reshape(shp)
            rcv = np.flip(rcv, 0)
            gcv = arr0[:,1].reshape(shp)
            gcv = np.flip(gcv, 0)
            bcv = arr0[:,2].reshape(shp)
            bcv = np.flip(bcv, 0)
            arr = np.array([rcv, gcv, bcv])
            arr = np.swapaxes(arr, 0, 2)
            arr = np.swapaxes(arr, 0, 1)
            vimg = Picture(arr)
            _loadcommon(vimg, d)
            objs.append(vimg)

        ### Text2D
        elif 'text2d' == d['type'].lower():
            t = vedo.shapes.Text2D(d['text'], font=d['font'], c=d['color'])
            t.pos(d['position']).size(d['size'])
            t.background(d['bgcol'], d['alpha'])
            if d['frame']:
                t.frame(d['bgcol'])
            objs.append(t)

        ### Annotation ## backward compatibility - will disappear
        elif 'annotation' == d['type'].lower():
            from vedo.shapes import Text2D
            pos = d['position']
            if isinstance(pos, int):
                pos = "top-left"
                d['size'] *= 2.7
            t = Text2D(d['text'], font=d['font'], c=d['color']).pos(pos)
            t.background(d['bgcol'], d['alpha']).size(d['size']).frame(d['bgcol'])
            objs.append(t) ## backward compatibility

    if len(objs) == 1:
        return objs[0]
    elif len(objs) == 0:
        return None
    else:
        return objs


def loadImageData(filename):
    """Read and return a ``vtkImageData`` object from file.
    Use ``load`` instead.
    E.g. `img = load('myfile.tif').imagedata()`
    """
    if ".tif" in filename.lower():
        reader = vtk.vtkTIFFReader()
    elif ".slc" in filename.lower():
        reader = vtk.vtkSLCReader()
        if not reader.CanReadFile(filename):
            colors.printc("\prohibited Sorry bad slc file " + filename, c='r')
            return None
    elif ".vti" in filename.lower():
        reader = vtk.vtkXMLImageDataReader()
    elif ".mhd" in filename.lower():
        reader = vtk.vtkMetaImageReader()
    elif ".dem" in filename.lower():
        reader = vtk.vtkDEMReader()
    elif ".nii" in filename.lower():
        reader = vtk.vtkNIFTIImageReader()
    elif ".nrrd" in filename.lower():
        reader = vtk.vtkNrrdReader()
        if not reader.CanReadFile(filename):
            colors.printc("\prohibited Sorry bad nrrd file " + filename, c='r')
            return None
    reader.SetFileName(filename)
    reader.Update()
    image = reader.GetOutput()
    return image



###########################################################
def write(objct, fileoutput, binary=True):
    """
    Write 3D object to file. (same as `save()`).

    Possile extensions are:
        - vtk, vti, npy, npz, ply, obj, stl, byu, vtp, vti, mhd, xyz, tif, png, bmp.
    """
    obj = objct
    if isinstance(obj, Points): # picks transformation
        obj = objct.polydata(True)
    elif isinstance(obj, (vtk.vtkActor, vtk.vtkVolume)):
        obj = objct.GetMapper().GetInput()
    elif isinstance(obj, (vtk.vtkPolyData, vtk.vtkImageData)):
        obj = objct

    if hasattr(obj, 'filename'):
        obj.filename = fileoutput

    fr = fileoutput.lower()
    if   fr.endswith(".vtk"):
        writer = vtk.vtkDataSetWriter()
    elif fr.endswith(".ply"):
        writer = vtk.vtkPLYWriter()
        writer.AddComment("PLY file generated by vedo")
        lut = objct.GetMapper().GetLookupTable()
        if lut:
            pscal = obj.GetPointData().GetScalars()
            if not pscal:
                pscal = obj.GetCellData().GetScalars()
            if pscal and pscal.GetName():
                writer.SetArrayName(pscal.GetName())
            writer.SetLookupTable(lut)
    elif fr.endswith(".stl"):
        writer = vtk.vtkSTLWriter()
    elif fr.endswith(".vtp"):
        writer = vtk.vtkXMLPolyDataWriter()
    elif fr.endswith(".vtu"):
        writer = vtk.vtkXMLUnstructuredGridWriter()
    elif fr.endswith(".vtm"):
        g = vtk.vtkMultiBlockDataGroupFilter()
        for ob in objct:
            if isinstance(ob, (Points, Volume)): # picks transformation
                ob = ob.polydata(True)
                g.AddInputData(ob)
            # elif isinstance(ob, (vtk.vtkActor, vtk.vtkVolume)):
            #     ob = ob.GetMapper().GetInput()
            #     g.AddInputData(ob)
        g.Update()
        mb = g.GetOutputDataObject(0)
        wri = vtk.vtkXMLMultiBlockDataWriter()
        wri.SetInputData(mb)
        wri.SetFileName(fileoutput)
        wri.Write()
        return mb
    elif fr.endswith(".xyz"):
        writer = vtk.vtkSimplePointsWriter()
    elif fr.endswith(".facet"):
        writer = vtk.vtkFacetWriter()
    elif fr.endswith(".tif"):
        writer = vtk.vtkTIFFWriter()
        writer.SetFileDimensionality(len(obj.GetDimensions()))
    elif fr.endswith(".vti"):
        writer = vtk.vtkXMLImageDataWriter()
    elif fr.endswith(".mhd"):
        writer = vtk.vtkMetaImageWriter()
    elif fr.endswith(".nii"):
        writer = vtk.vtkNIFTIImageWriter()
    elif fr.endswith(".png"):
        writer = vtk.vtkPNGWriter()
    elif fr.endswith(".jpg"):
        writer = vtk.vtkJPEGWriter()
    elif fr.endswith(".bmp"):
        writer = vtk.vtkBMPWriter()
    elif fr.endswith(".npy") or fr.endswith(".npz"):
        if utils.isSequence(objct):
            objslist = objct
        else:
            objslist = [objct]
        dicts2save = []
        for obj in objslist:
            dicts2save.append( toNumpy(obj) )
        np.save(fileoutput, dicts2save)
        return dicts2save

    elif fr.endswith(".obj"):
        outF = open(fileoutput, "w")
        outF.write('# OBJ file format with ext .obj\n')
        outF.write('# File generated by vedo\n')

        for p in objct.points():
            outF.write("v {:.5g} {:.5g} {:.5g}\n".format(*p))

        # pdata = objct.polydata().GetPointData().GetScalars()
        # if pdata:
        #     ndata = vtk_to_numpy(pdata)
        #     for vd in ndata:
        #         outF.write('vp '+ str(vd) +'\n')

        #ptxt = objct.polydata().GetPointData().GetTCoords() # not working
        #if ptxt:
        #    ntxt = vtk_to_numpy(ptxt)
        #    print(len(objct.faces()), objct.points().shape, ntxt.shape)
        #    for vt in ntxt:
        #        outF.write('vt '+ str(vt[0]) +" "+ str(vt[1])+ ' 0\n')

        for i,f in enumerate(objct.faces()):
            fs = ''
            for fi in f:
                fs += " {:d}".format(fi+1)
            outF.write('f' + fs + '\n')

        for l in objct.lines():
            ls = ''
            for li in l:
                ls += str(li+1)+" "
            outF.write('l '+ ls + '\n')

        outF.close()
        return objct

    elif fr.endswith(".xml"):  # write tetrahedral dolfin xml
        vertices = objct.points().astype(str)
        faces = np.array(objct.faces()).astype(str)
        ncoords = vertices.shape[0]
        outF = open(fileoutput, "w")
        outF.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        outF.write('<dolfin xmlns:dolfin="http://www.fenicsproject.org">\n')

        if len(faces[0]) == 4:# write tetrahedral mesh
            ntets = faces.shape[0]
            outF.write('  <mesh celltype="tetrahedron" dim="3">\n')
            outF.write('    <vertices size="' + str(ncoords) + '">\n')
            for i in range(ncoords):
                x, y, z = vertices[i]
                outF.write('      <vertex index="'+str(i)+'" x="'+x+'" y="'+y+'" z="'+z+'"/>\n')
            outF.write('    </vertices>\n')
            outF.write('    <cells size="' + str(ntets) + '">\n')
            for i in range(ntets):
                v0, v1, v2, v3 = faces[i]
                outF.write('     <tetrahedron index="'+str(i)
                           + '" v0="'+v0+'" v1="'+v1+'" v2="'+v2+'" v3="'+v3+'"/>\n')

        elif len(faces[0]) == 3:# write triangle mesh
            ntri = faces.shape[0]
            outF.write('  <mesh celltype="triangle" dim="2">\n')
            outF.write('    <vertices size="' + str(ncoords) + '">\n')
            for i in range(ncoords):
                x, y, dummy_z = vertices[i]
                outF.write('      <vertex index="'+str(i)+'" x="'+x+'" y="'+y+'"/>\n')
            outF.write('    </vertices>\n')
            outF.write('    <cells size="' + str(ntri) + '">\n')
            for i in range(ntri):
                v0, v1, v2 = faces[i]
                outF.write('     <triangle index="'+str(i)+'" v0="'+v0+'" v1="'+v1+'" v2="'+v2+'"/>\n')

        outF.write('    </cells>\n')
        outF.write("  </mesh>\n")
        outF.write("</dolfin>\n")
        outF.close()
        return objct

    else:
        colors.printc("\noentry Unknown format", fileoutput, "file not saved.", c="r")
        return objct

    try:
        if hasattr(writer, 'SetFileTypeToBinary'):
            if binary:
                writer.SetFileTypeToBinary()
            else:
                writer.SetFileTypeToASCII()
        writer.SetInputData(obj)
        writer.SetFileName(fileoutput)
        writer.Write()
    except Exception as e:
        colors.printc("\noentry Error saving: " + fileoutput, "\n", e, c="r")
    return objct


def writeTransform(inobj, filename='transform.mat', comment=''):
    """
    Save a transformation for a mesh or pointcloud to ASCII file.

    Parameters
    ----------
    filename : str, optional
        output file name. The default is 'transform.mat'.
    comment : str, optional
        some optional comment. The default is ''.

    Returns
    -------
    None.
    """
    if isinstance(inobj, Points):
        M = inobj.getTransform().GetMatrix()
    elif isinstance(inobj, vtk.vtkTransform):
        M = inobj.GetMatrix()
    elif isinstance(inobj, vtk.vtkMatrix4x4):
        M = inobj
    else:
        colors.printc("Error in io.writeTransform: cannot understand input type",
                      type(inobj), c='r')
    with open(filename,'w') as f:
        if comment:
            f.write('#'+comment+'\n')
        for i in range(4):
            f.write( str(M.GetElement(i,0))+' '+
                     str(M.GetElement(i,1))+' '+
                     str(M.GetElement(i,2))+' '+
                     str(M.GetElement(i,3))+'\n',
                    )
        f.write('\n')
    return

def loadTransform(filename):
    """
    Load a ``vtkTransform`` from a file.mat.

    Returns
    -------
    T : vtkTransform
        The transformation to be applied to some object (``use applyTransform()``).
    comment : str
        a comment string associated to this transformation file.
    """
    with open(filename,'r') as f:
        lines = f.readlines()
        M = vtk.vtkMatrix4x4()
        i=0
        comment = ''
        for l in lines:
            if l.startswith('#'):
                comment = l.replace('#', "").replace('\n', "")
                continue
            vals = l.split(' ')
            if len(vals)==4:
                for j in range(4):
                    v = vals[j].replace('\n', '')
                    M.SetElement(i,j, float(v))
                i+=1
        T = vtk.vtkTransform()
        T.SetMatrix(M)
    return (T, comment)


###############################################################################
def exportWindow(fileoutput, binary=False):
    '''
    Exporter which writes out the renderered scene into an HTML, X3D
    or Numpy file.

    |export_x3d| |export_x3d.py|_

        `generated webpage <https://vedo.embl.es/examples/embryo.html>`_

        See also: FEniCS test `webpage <https://vedo.embl.es/examples/fenics_elasticity.html>`_.

    .. note:: the rendering window can also be exported to `numpy` file `scene.npz`
        by pressing ``E`` keyboard at any moment during visualization.
    '''
    fr = fileoutput.lower()

    ####################################################################
    if fr.endswith(".npy") or fr.endswith(".npz"):
        sdict = dict()
        vp = settings.plotter_instance
        sdict['shape'] = vp.shape #todo
        sdict['sharecam'] = vp.sharecam #todo
        sdict['camera'] = dict( pos=vp.camera.GetPosition(),
                                focalPoint=vp.camera.GetFocalPoint(),
                                viewup=vp.camera.GetViewUp(),
                                distance=vp.camera.GetDistance(),
                                clippingRange=vp.camera.GetClippingRange() )
        sdict['position'] = vp.pos
        sdict['size'] = vp.size
        sdict['axes'] = vp.axes
        sdict['title'] = vp.title
        sdict['xtitle'] = vp.xtitle
        sdict['ytitle'] = vp.ytitle
        sdict['ztitle'] = vp.ztitle
        sdict['backgrcol'] = colors.getColor(vp.backgrcol)
        sdict['backgrcol2'] = None
        if vp.renderer.GetGradientBackground():
            sdict['backgrcol2'] = vp.renderer.GetBackground2()
        sdict['useDepthPeeling'] = settings.useDepthPeeling
        sdict['renderLinesAsTubes'] = settings.renderLinesAsTubes
        sdict['hiddenLineRemoval'] = settings.hiddenLineRemoval
        sdict['visibleGridEdges'] = settings.visibleGridEdges
        sdict['interactorStyle'] = settings.interactorStyle
        sdict['useParallelProjection'] = settings.useParallelProjection
        sdict['defaultFont'] = settings.defaultFont
        sdict['objects'] = []

        allobjs = vp.getMeshes(includeNonPickables=True) + vp.getVolumes(includeNonPickables=True)
        acts2d = vp.renderer.GetActors2D()
        acts2d.InitTraversal()
        for i in range(acts2d.GetNumberOfItems()):
            a = acts2d.GetNextItem()
            if isinstance(a, vedo.Text2D):
                allobjs.append(a)

        allobjs = list(set(allobjs)) # make sure its unique

        for a in allobjs:
            sdict['objects'].append(toNumpy(a))

        if fr.endswith(".npz"):
            np.savez_compressed(fileoutput, vedo_scenes=[sdict])
        else:
            np.save(fileoutput, [sdict])

    ####################################################################
    # elif fr.endswith(".obj"):
    #     w = vtk.vtkOBJExporter()
    #     w.SetInputData(settings.plotter_instance.window)
    #     w.Update()
    #     colors.printc("\save Saved file:", fileoutput, c="g")


    ####################################################################
    elif fr.endswith(".x3d"):
        from vedo.docs import x3d_html
        exporter = vtk.vtkX3DExporter()
        exporter.SetBinary(binary)
        exporter.FastestOff()
        exporter.SetInput(settings.plotter_instance.window)
        exporter.SetFileName(fileoutput)
        exporter.Update()
        exporter.Write()
        x3d_html = x3d_html.replace("~fileoutput", fileoutput)
        wsize = settings.plotter_instance.window.GetSize()
        x3d_html = x3d_html.replace("~width", str(wsize[0]))
        x3d_html = x3d_html.replace("~height", str(wsize[1]))
        outF = open(fileoutput.replace('.x3d', '.html'), "w")
        outF.write(x3d_html)
        outF.close()
        colors.printc("\save Saved files:", fileoutput,
                      fileoutput.replace('.x3d', '.html'), c="g")

    ####################################################################
    elif fr.endswith(".html"):
        from vedo.backends import getNotebookBackend

        savebk = settings.notebookBackend
        settings.notebookBackend='k3d'
        plt = getNotebookBackend(settings.plotter_instance.actors, 1.5, '')

        with open(fileoutput,'w') as fp:
            fp.write(plt.get_snapshot())

        settings.notebookBackend = savebk

    else:
        colors.printc("Export extension", fr.split('.')[-1],
                      "is not supported.", c='r')
    return


def importWindow(fileinput, mtlFile=None, texturePath=None):
    """Import a whole scene from a Numpy or OBJ wavefront file.
    Return a ``Plotter`` instance.

    :param str mtlFile: MTL file for OBJ wavefront files.
    :param str texturePath: path of the texture files directory.
    """
    from vedo import Plotter

    data = None
    if isinstance(fileinput, dict):
        data = fileinput
    elif fileinput.endswith('.npy'):
        data = np.load(fileinput, allow_pickle=True, encoding="latin1").flatten()[0]
    elif fileinput.endswith('.npz'):
        data = np.load(fileinput, allow_pickle=True)['vedo_scenes'][0]

    if data is not None:
        if 'renderLinesAsTubes' in data.keys():
            settings.renderLinesAsTubes = data['renderLinesAsTubes']
        if 'hiddenLineRemoval' in data.keys():
            settings.hiddenLineRemoval = data['hiddenLineRemoval']
        if 'visibleGridEdges' in data.keys():
            settings.visibleGridEdges = data['visibleGridEdges']
        if 'interactorStyle' in data.keys():
            settings.interactorStyle = data['interactorStyle']
        if 'useParallelProjection' in data.keys():
            settings.useParallelProjection = data['useParallelProjection']
        if 'usePolygonOffset' in data.keys():
            settings.usePolygonOffset = data['usePolygonOffset']
        if 'polygonOffsetFactor' in data.keys():
            settings.polygonOffsetFactor = data['polygonOffsetFactor']
        if 'polygonOffsetUnits' in data.keys():
            settings.polygonOffsetUnits = data['polygonOffsetUnits']
        if 'interpolateScalarsBeforeMapping' in data.keys():
            settings.interpolateScalarsBeforeMapping = data['interpolateScalarsBeforeMapping']
        if 'defaultFont' in data.keys():
            settings.defaultFont = data['defaultFont']

        axes = data.pop('axes', 4)
        title = data.pop('title', '')
        backgrcol = data.pop('backgrcol', "white")
        backgrcol2 = data.pop('backgrcol2', None)
        cam = data.pop('camera', None)

        if data['shape'] != (1,1): data['size']="auto" # disable size

        vp = Plotter(size=data['size'], # not necessarily a good idea to set it
                     #shape=data['shape'], # will need to create a Renderer class first
                     axes=axes,
                     title=title,
                     bg=backgrcol,
                     bg2=backgrcol2,
        )
        vp.xtitle = data.pop('xtitle', 'x')
        vp.ytitle = data.pop('ytitle', 'y')
        vp.ztitle = data.pop('ztitle', 'z')

        if cam:
            if 'pos' in cam.keys(): vp.camera.SetPosition( cam['pos'] )
            if 'focalPoint' in cam.keys(): vp.camera.SetFocalPoint( cam['focalPoint'] )
            if 'viewup' in cam.keys(): vp.camera.SetViewUp(cam['viewup']  )
            if 'distance' in cam.keys(): vp.camera.SetDistance( cam['distance'] )
            if 'clippingRange' in cam.keys(): vp.camera.SetClippingRange( cam['clippingRange'] )
            vp.resetcam = False

        if 'objects' in data.keys():
            objs = loadNumpy(data['objects'])
            if not utils.isSequence(objs):
               objs = [objs]
        else:
            #colors.printc("Trying to import a single mesh.. use load() instead.", c='r')
            #colors.printc(" -> try to load a single object with load().", c='r')
            objs = [loadNumpy(fileinput)]

        vp.actors = objs
        return vp

    elif '.obj' in fileinput.lower():

        vp = Plotter()

        importer = vtk.vtkOBJImporter()
        importer.SetFileName(fileinput)
        if mtlFile is not False:
            if mtlFile is None:
                mtlFile = fileinput.replace('.obj', '.mtl').replace('.OBJ', '.MTL')
            importer.SetFileNameMTL(mtlFile)
        if texturePath is not False:
            if texturePath is None:
                texturePath = fileinput.replace('.obj', '.txt').replace('.OBJ', '.TXT')
            importer.SetTexturePath(texturePath)
        importer.SetRenderWindow(vp.window)
        importer.Update()

        actors = vp.renderer.GetActors()
        actors.InitTraversal()
        for i in range(actors.GetNumberOfItems()):
            vactor = actors.GetNextActor()
            act = Mesh(vactor)
            act_tu = vactor.GetTexture()
            if act_tu:
                act_tu.InterpolateOn()
                act.texture(act_tu)
            vp.actors.append( act )
        return vp


##########################################################
def screenshot(filename="screenshot.png", scale=None, returnNumpy=False):
    """
    Save a screenshot of the current rendering window.

    :param int scale: set image magnification
    :param bool returnNumpy: return a numpy array of the image
    """
    if not settings.plotter_instance or not settings.plotter_instance.window:
        colors.printc('\bomb screenshot(): Rendering window is not present, skip.', c='r')
        return settings.plotter_instance

    if filename.endswith('.pdf'):
        writer = vtk.vtkGL2PSExporter()
        writer.SetRenderWindow(settings.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToPDF()
        writer.SetFilePrefix(filename.replace('.pdf',''))
        writer.Write()
        return settings.plotter_instance ##########
    elif filename.endswith('.svg'):
        writer = vtk.vtkGL2PSExporter()
        writer.SetRenderWindow(settings.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToSVG()
        writer.SetFilePrefix(filename.replace('.svg',''))
        writer.Write()
        return settings.plotter_instance ##########
    elif filename.endswith('.eps'):
        writer = vtk.vtkGL2PSExporter()
        writer.SetRenderWindow(settings.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToEPS()
        writer.SetFilePrefix(filename.replace('.eps',''))
        writer.Write()
        return settings.plotter_instance ##########

    if scale is None:
        scale = settings.screeshotScale

    if settings.screeshotLargeImage:
       w2if = vtk.vtkRenderLargeImage()
       w2if.SetInput(settings.plotter_instance.renderer)
       w2if.SetMagnification(scale)
    else:
        w2if = vtk.vtkWindowToImageFilter()
        w2if.SetInput(settings.plotter_instance.window)
        if hasattr(w2if, 'SetScale'):
            w2if.SetScale(scale, scale)
        if settings.screenshotTransparentBackground:
            w2if.SetInputBufferTypeToRGBA()
        w2if.ReadFrontBufferOff()  # read from the back buffer
    w2if.Update()

    if returnNumpy:
        w2ifout = w2if.GetOutput()
        npdata = vtk_to_numpy(w2ifout.GetPointData().GetArray("ImageScalars"))
        npdata = npdata[:,[0,1,2]]
        ydim, xdim, _ = w2ifout.GetDimensions()
        npdata = npdata.reshape([xdim, ydim, -1])
        npdata = np.flip(npdata, axis=0)
        return npdata

    if filename.lower().endswith('.png'):
        writer = vtk.vtkPNGWriter()
        writer.SetFileName(filename)
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    elif filename.lower().endswith('.jpg') or filename.lower().endswith('.jpeg'):
        writer = vtk.vtkJPEGWriter()
        writer.SetFileName(filename)
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    else: #add .png
        writer = vtk.vtkPNGWriter()
        writer.SetFileName(filename+'.png')
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    return settings.plotter_instance

def ask(*question, **kwarg):
    """
    Ask a question from command line. Return the answer as a string.
    See function `printc()` for the description of the options.
    """
    kwarg.update({'end': ' '})
    if 'invert' not in kwarg.keys():
        kwarg.update({'invert': True})
    if 'box' in kwarg.keys():
        kwarg.update({'box': ''})
    colors.printc(*question, **kwarg)
    resp = input()
    return resp

class Video:
    """
    Class to generate a video from the specified rendering window.
    Program ``ffmpeg`` is used to create video from each generated frame.
    :param str name: name of the output file.
    :param int fps: set the number of frames per second.
    :param float duration: set the total `duration` of the video and recalculates `fps` accordingly.
    :param str ffmpeg: set path to ffmpeg program. Default value assumes ffmpeg command is in the path.

    |makeVideo| |makeVideo.py|_
    """

    def __init__(self,
                 name="movie.mp4",
                 duration=None,
                 fps=24,
                 backend='ffmpeg',
                ):

        from tempfile import TemporaryDirectory

        self.name = name
        self.duration = duration
        self.backend = backend
        self.fps = float(fps)
        self.command = "ffmpeg -loglevel panic -y -r"
        self.options = "-b:v 8000k"

        self.frames = []
        self.tmp_dir = TemporaryDirectory()
        self.get_filename = lambda x: os.path.join(self.tmp_dir.name, x)
        colors.printc("\video Video", self.name, "is open...", c="m")

    def addFrame(self):
        """Add frame to current video."""
        fr = self.get_filename(str(len(self.frames)) + ".png")
        screenshot(fr)
        self.frames.append(fr)
        return self

    def pause(self, pause=0):
        """Insert a `pause`, in seconds."""
        fr = self.frames[-1]
        n = int(self.fps * pause)
        for _ in range(n):
            fr2 = self.get_filename(str(len(self.frames)) + ".png")
            self.frames.append(fr2)
            os.system("cp -f %s %s" % (fr, fr2))
        return self


    def action(self, elevation_range=(0,80),
               azimuth_range=(0,359),
               zoom=None,
               cam1=None, cam2=None,
               resetcam=False,
               ):
        """
        Automatic shooting of a static scene by specifying rotation and elevation ranges.

        :param list elevation_range: initial and final elevation angles
        :param list azimuth_range: initial and final azimuth angles
        :param float zoom: initial zooming
        :param cam1 cam2: initial and final camera position, can be dictionary or a vtkCamera
        """
        if not self.duration:
            self.duration = 5

        def buildcam(cm):
            cm_pos = cm.pop("pos", None)
            cm_focalPoint = cm.pop("focalPoint", None)
            cm_viewup = cm.pop("viewup", None)
            cm_distance = cm.pop("distance", None)
            cm_clippingRange = cm.pop("clippingRange", None)
            cm_parallelScale = cm.pop("parallelScale", None)
            cm_thickness = cm.pop("thickness", None)
            cm_viewAngle = cm.pop("viewAngle", None)
            cm = vtk.vtkCamera()
            if cm_pos is not None: cm.SetPosition(cm_pos)
            if cm_focalPoint is not None: cm.SetFocalPoint(cm_focalPoint)
            if cm_viewup is not None: cm.SetViewUp(cm_viewup)
            if cm_distance is not None: cm.SetDistance(cm_distance)
            if cm_clippingRange is not None: cm.SetClippingRange(cm_clippingRange)
            if cm_parallelScale is not None: cm.SetParallelScale(cm_parallelScale)
            if cm_thickness is not None: cm.SetThickness(cm_thickness)
            if cm_viewAngle is not None: cm.SetViewAngle(cm_viewAngle)
            return cm

        vp = settings.plotter_instance

        if zoom:
            vp.camera.Zoom(zoom)

        if isinstance(cam1, dict):
            cam1 = buildcam(cam1)
        if isinstance(cam2, dict):
            cam2 = buildcam(cam2)

        if len(elevation_range)==2:
            vp.camera.Elevation(elevation_range[0])
        if len(azimuth_range)==2:
            vp.camera.Azimuth(azimuth_range[0])

        vp.show(resetcam=resetcam, interactive=False)
        # if resetcam: vp.renderer.ResetCamera()

        n = self.fps * self.duration
        for i in range(int(n)):
            if cam1 and cam2:
                vp.moveCamera(cam1, cam2, i/n)
            else:
                if len(elevation_range)==2:
                    vp.camera.Elevation((elevation_range[1]-elevation_range[0])/n)
                if len(azimuth_range)==2:
                    vp.camera.Azimuth((azimuth_range[1]-azimuth_range[0])/n)
            vp.show()
            self.addFrame()
        return self

    def close(self):
        """
        Render the video and write to file.
        Return the current Plotter instance.
        """
        if self.duration:
            self.fps = len(self.frames) / float(self.duration)
            colors.printc("Recalculated video FPS to", round(self.fps, 3), c="m")
        else:
            self.fps = int(self.fps)

        self.name = self.name.split('.')[0]+'.mp4'

        ########################################
        if self.backend == 'ffmpeg':
            out = os.system(self.command + " " + str(self.fps)
                            + " -i " + self.tmp_dir.name + os.sep
                            + "%01d.png " + self.options + " " + self.name)
            if out:
                colors.printc("ffmpeg returning error", c='r')
            else:
                colors.printc("\save Video saved as", self.name, c="m")

        ########################################
        elif 'cv' in self.backend:
            try:
                import cv2
            except:
                colors.printc("Error in Video backend: opencv not installed!", c='r')
                return

            cap = cv2.VideoCapture(os.path.join(self.tmp_dir.name, "%1d.png"))
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            w,h = settings.plotter_instance.window.GetSize()
            writer = cv2.VideoWriter(self.name, fourcc, self.fps, (w, h), True)

            found = False
            while True:
                ret, frame = cap.read()
                if not ret: break
                writer.write(frame)
                found = True

            cap.release()
            writer.release()
            if found:
                colors.printc("\save Video saved as", self.name, c="m")
            else:
                colors.printc("could not find snapshots", c='r')

        self.tmp_dir.cleanup()
        return settings.plotter_instance





"""Glyphs:
at each vertex of a mesh, another mesh
is shown with various orientation options
"""
from vedo import *
import numpy as np

s = Sphere(res=12).c('white',0.1).wireframe()

randvs = np.random.rand(s.NPoints(), 3)  # random orientation vectors

#######################################
gly1 = Ellipsoid().scale(0.04)

gsphere1 = Glyph(s, gly1,
                 orientationArray=randvs,
                 scaleByVectorSize=True,
                 colorByVectorSize=True,
                 c='jet',
                 )


#######################################
gly2 = load(dataurl+"shuttle.obj").rotateY(180).scale(0.02)

gsphere2 = Glyph(s, gly2,
                 orientationArray="normals",
                 tol=0.1,  # minimum seaparation of 10% of bounding box
                 c='lightblue',
                 )

# show two groups of objects on N=2 renderers:
show([(s, gsphere1, __doc__), (s, gsphere2)], N=2, bg='bb', zoom=1.4)



"""Mirror a mesh along one of the Cartesian axes.
(Hover mouse to identify original and mirrored)"""
from vedo import *

myted1 = load(dataurl+"teddy.vtk").flag('original')

myted2 = myted1.clone(deep=False).mirror("y")
myted2.pos(0,3,0).c("green").flag('mirrored')

show(myted1, myted2, __doc__, axes=2, viewup="z", bg2='ly')



"""Markers set, analogous to matplotlib"""
from vedo import *

symbols = ['.', 'p','*','h','D','d','o','v','>','<','s','x','+','a']

vp = Plotter(size=(1500,300), axes=0)
for i,s in enumerate(symbols):
    vp += Marker(s, filled=True).x(i*0.6).color(i)
    vp += Text3D(s, pos=[i*0.6,-0.6,0], s=0.12).color('k')
vp += __doc__

vp.show(zoom=5, viewup='2d')



"""Plot streamlines of the 2D field:

u(x,y) = -1 - x\^2 + y
v(x,y) =  1 + x  - y\^2
"""
from vedo import Points, show
from vedo.pyplot import streamplot
import numpy as np

# a grid with a vector field (U,V):
X, Y = np.mgrid[-5:5 :15j, -4:4 :15j]
U = -1 - X**2 + Y
V =  1 + X    - Y**2

# optionally, pick some random points as seeds:
prob_pts = np.random.rand(200, 2)*8 - [4,4]

sp = streamplot(X,Y, U,V,
                lw=0.001,            # line width in abs. units
                direction='forward', # 'both' or 'backward'
                probes=prob_pts,
               )

pts = Points(prob_pts, r=5, c='white')

show(sp, pts, __doc__, axes=1, bg='bb')



"""Generate a lineage graph
of cell divisions
"""
# N.B.: no positions are specified here, only connectivity!
from vedo import show, settings
from vedo.pyplot import DirectedGraph

settings.allowSubScripts=0

# Layouts: [2d, fast2d, clustering2d, circular, circular3d, cone, force, tree]
#g = Graph(layout='2d', zrange=7)
g = DirectedGraph(layout='cone')
#g = DirectedGraph(layout='circular3d', height=1, radius=1.5)
#g = DirectedGraph(layout='force')

# Vertex generation is automatic,
#  add a child to vertex0, so that now vertex1 exists
g.addChild(0, edgeLabel="Mother giving birth\nto her little baby cell")
g.addChild(1); g.addChild(1)
g.addChild(2); g.addChild(2); g.addChild(2)
g.addChild(3); g.addChild(3, edgeLabel="It's a male!")
g.addChild(4); g.addChild(4)
for i in range(7): g.addChild(5, nodeLabel="cell5_"+str(i))
g.addChild(7); g.addChild(7); g.addChild(7)

g.build() # optimize layout

g.unpack(0).color('dg').lineWidth(3) #0=graph, 1=vertexLabels, 2=edgeLabels, 3=arrows
g.unpack(2).color('dr')

show(g, __doc__, axes=9, elevation=-40)



"""Colorize faces of a Mesh
passing a 1-to-1 list of colors and
optionally a list of transparencies"""
from vedo import *

tor = Torus(res=9).lineWidth(0.1)

cols, alphas = [], []
for i in range(tor.NCells()):
    cols.append(i)                # i-th color
    alphas.append(i/tor.NCells()) # a transparency value

tor.cellIndividualColors(cols, alphas)
print('Mesh arrays:', tor.getArrayNames())

show(tor, __doc__)



"""
Example to show how to use recoSurface()
to reconstruct a surface from points.

 1. An object is loaded and
    noise is added to its vertices.
 2. the point cloud is smoothened with MLS
    (see moving_least_squares.py)
 3. mesh.clean() imposes a minimum distance
    among mesh points where 'tol' is the
    fraction of the mesh size.
 4. a triangular mesh is extracted from
    this set of sparse Points, 'bins' is the
    number of voxels of the subdivision
"""
print(__doc__)
from vedo import *
import numpy as np


vp = Plotter(N=4, axes=0)

mesh = vp.load(dataurl+"apple.ply").subdivide()
vp.show(mesh, at=0)

noise = np.random.randn(mesh.N(), 3) * 0.03

pts0 = Points(mesh.points() + noise, r=3).legend("noisy cloud")
vp.show(pts0, at=1)

pts1 = pts0.clone().smoothMLS2D(f=0.8)  # smooth cloud

print("Nr of points before cleaning nr. points:", pts1.N())

# impose a min distance among mesh points
pts1.clean(tol=0.005).legend("smooth cloud")
print("             after  cleaning nr. points:", pts1.N())

vp.show(pts1, at=2)

# reconstructed surface from point cloud
reco = recoSurface(pts1, dims=100, radius=0.2).legend("surf. reco")
vp.show(reco, at=3, axes=7, zoom=1.2, interactive=1)



"""Customizing Axes.
Cartesian planes can be displaced
from their lower-range default position"""
from vedo import Sphere, Axes, precision, show

sph = Sphere().scale([4,3,2]).shift(5,6,7).c('green2', 0.1)

axs = Axes(sph,          # build axes for object sph
           xtitle="x axis",
           ytitle="y axis",
           ztitle="z axis",
           htitle='An ellipsoid at '+precision(sph.centerOfMass(),2),
           hTitleFont=1,
           hTitleColor='red3',
           zxGrid=True,
           xyFrameLine=2, yzFrameLine=2, zxFrameLine=2,
           xyFrameColor='red3',
           yzFrameColor='green3',
           zxFrameColor='blue3',
           xyShift=0.2,  # move xy plane 20% along z
           yzShift=0.2,  # move yz plane 20% along x
           zxShift=0.2,  # move zx plane 20% along y
          )

show(sph, axs, __doc__)



"""Use iminuit to find the minimum of a 3D scalar field"""
from vedo import show, Point, Line, printc
from iminuit import Minuit
# pip install iminuit  # https://github.com/scikit-hep/iminuit
import numpy as np


def fcn(x, y, z):
    f =  (x - 4) ** 4 + (y - 3) ** 4 + (z - 2) ** 2
    if not vals or f < vals[-1]:
        path.append([x,y,z])
        vals.append(f)
    return f


paths = []
for x,y,z in  np.random.rand(200, 3)*3:
    path, vals = [], []
    m = Minuit(fcn, x=x, y=y, z=z)
    m.errordef = m.LEAST_SQUARES

    # m.simplex()  # run simplex optimiser
    m.migrad()     # run migrad optimiser

    line = Line(path).cmap('jet_r', vals).lw(2).alpha(0.25)
    paths.append(line)

printc('Last optimization output:', c='green7', invert=1)
printc(m, c='green7', italic=1)
show(paths, Point([4,3,2]), __doc__, axes=1)




"""Generate a polygonal Mesh
from a contour line."""
from vedo import *
from vedo.pyplot import histogram

shapes = load(dataurl+'timecourse1d.npy') # list of lines
shape = shapes[56].mirror().rotateZ(-90)

msh = shape.tomesh()       # Generate the Mesh from the line
msh.smoothLaplacian()      # make the triangles more uniform
msh.addQuality().addScalarBar3D() # Measure triangle quality

contour = Line(shape).c('red4').lw(5)
histo = histogram(msh.getCellArray('Quality'), aspect=3/4,
                  c='RdYlBu', xtitle='triangle mesh quality')
show([
      [contour, contour.labels('id'), msh, __doc__],
      histo,
     ], N=2, sharecam=False,
)


"""
Simulation of bacteria types that divide at a given rate
As they divide they occupy more and more space
"""
print(__doc__)
from vedo import Plotter, ProgressBar, pcaEllipsoid, Points, Line
import numpy as np


##############################################################
class Colony:
    def __init__(self, cells=(), c="b", cellSize=5):
        self.cells = cells
        self.cellSize = cellSize
        self.color = c


##############################################################
class Cell:
    def __init__(self, pos=(0, 0, 0), tdiv=10):

        self.pos = pos  # position in space

        self.tdiv = tdiv  # after this time cell will divide on average
        self.tdiv_spread = 2  # gaussian spread to determine division time
        self.lag = 2  # cell division cannot happen before this age

        self.split_dist = 0.2  # split at this initial distance
        self.split_dist_spread = 0.05  # gauss split initial distance spread

        self.apoptosis = 90  # after this time cell dies
        self.apoptosis_spread = 1

        self.t = 0  # current absolute clock time
        self.timeOfBirth = 0  # time of life since birth
        self.rndwalk = 0.001  # random walk step

        # decide random life and death times for this specific cell
        self.celltdiv = self.tdiv + np.random.randn() * self.tdiv_spread
        self.celltdiv = max(self.lag, self.celltdiv)
        self.apoptosis += np.random.randn() * self.apoptosis_spread
        self.apoptosis = max(self.lag, self.apoptosis)

    def divideAt(self, t):
        self.t = t
        self.pos += np.random.normal(size=3) * self.rndwalk
        if t > self.timeOfBirth + self.celltdiv:
            return True
        else:
            return False

    def dieAt(self, t):
        if t > self.timeOfBirth + self.apoptosis:
            return True
        else:
            return False

    def split(self):
        c = Cell(self.pos, tdiv=self.tdiv)
        # this generates a random point uniformly on a sphere
        p = np.random.normal(size=3)
        radius = self.split_dist + np.random.normal() * self.split_dist_spread
        ps = np.linalg.norm(p) * 2 / radius
        d = p / ps
        c.pos = self.pos + d
        self.pos = self.pos - d
        self.timeOfBirth = self.t
        c.timeOfBirth = self.t
        return c

    def dist(self, c):
        v = self.pos - np.array(c.pos)
        s = np.linalg.norm(v)
        return v, s


##############################################################################
vp = Plotter(interactive=0, axes=3)

# place vtkCamera at a specific position
# (get these numbers by pressing Shift-C)
vp.camera.SetPosition([2.5, 2.5, 5.5])
vp.camera.SetFocalPoint([0.4, 0.4, 0.4])
vp.camera.SetParallelScale(1.8)
vp.camera.SetViewUp([-0.1, 1, -0.3])

# Let's start with creating 3 colonies of 1 cell each
# of types: red, green and blue, in different positions in space
# and with 3 different rates of division (tdiv in hours)
c1 = Colony([Cell([1, 0, 0], tdiv=8)], c="b")
c2 = Colony([Cell([0, 1, 0], tdiv=9)], c="g")
c3 = Colony([Cell([0, 0, 1], tdiv=10)], c="r")
colonies = [c1, c2, c3]

# time goes from 0 to 90 hours
pb = ProgressBar(0, 50, step=0.1, c=1)
for t in pb.range():
    msg = "[Nb,Ng,Nr,t] = "
    vp.actors = []  # clean up the list of actors

    for colony in colonies:

        newcells = []
        for cell in colony.cells:

            if cell.dieAt(t):
                continue
            if cell.divideAt(t):
                newc = cell.split()  # make daughter cell
                vp += Line(cell.pos, newc.pos, c="k", lw=3, alpha=0.5)
                newcells.append(newc)
            newcells.append(cell)
        colony.cells = newcells

        pts = [c.pos for c in newcells]  # draw all points at once
        vp += Points(pts, c=colony.color, r=5, alpha=0.80)   # nucleus
        vp += Points(pts, c=colony.color, r=15, alpha=0.05)  # halo
        msg += str(len(colony.cells)) + ","

    pb.print(msg + str(int(t)))
    vp.show(resetcam=0)

# draw the oriented ellipsoid that contains 50% of the cells
for colony in colonies:
    pts = [c.pos for c in colony.cells]
    a = pcaEllipsoid(pts, pvalue=0.5)
    a.color(colony.color).alpha(0.3)
    a.legend("1/rate=" + str(colony.cells[0].tdiv) + "h")
    vp += a

vp.show(resetcam=0, interactive=1)



"""Use of plot() function analogous to matplotlib"""
import numpy as np, vtk
from vedo import *
from vedo.pyplot import plot

x = np.linspace(0, 5, 10)

plt1 = plot(x, x*x,    'sg-',  title='Plot1: y=x*x')
plt2 = plot(x, cos(x), 'pr--', title='Plot2: y=cos(x)')
plt3 = plot(x, sqrt(x),'Db-',  title='Plot3: y=sqrt(x)')
plt4 = plot(x, sin(x), '*t--', title='Plot4: y=sin(x)')

# window shape can be expressed as "n/m" or "n|m"
show(plt1, plt2, plt3, plt4,
     shape="3|1", sharecam=False, size=(1300,900), interactive=True)

printc('plt1 is vtkAssembly?', isinstance(plt1, vtk.vtkAssembly))


"""Use two sliders to change
color and transparency of a mesh"""
from vedo import Plotter, dataurl


def slider1(widget, event):
    value = widget.GetRepresentation().GetValue()
    mesh.color(value)

def slider2(widget, event):
    value = widget.GetRepresentation().GetValue()
    mesh.alpha(value)


vp = Plotter(axes=0)
mesh = vp.load(dataurl+"magnolia.vtk").flat().lw(0.1)

vp.addSlider2D(slider1, -9, 9, value=0,
               pos="bottom-right", title="color number")

vp.addSlider2D(slider2, xmin=0.01, xmax=0.99, value=0.5, c="blue",
                pos="bottom-right-vertical", title="alpha value (opacity)")

vp += __doc__
vp.show()



"""Use array to vary radius and color
of a line represented as a tube.
"""
from vedo import *
import numpy as np

settings.defaultFont = 'Quikhand'

ln = [[sin(x), cos(x), x / 2] for x in np.arange(0,9, 0.1)]
N = len(ln)

############################### a simple tube( along ln
t1 = Tube(ln, c="blue", r=0.08)

############################### vary radius
rads = [0.3*(cos(6.0*ir/N))**2+0.1 for ir in range(N)]
t2 = Tube(ln, r=rads, c="tomato", res=24)

############################### vary color
cols = [i for i in range(N)]
cols = makeBands(cols, 5)  # make color bins
t3 = Tube(ln, r=rads, c=cols, res=24)


show(t1, __doc__, at=0, N=3, axes=dict(textScale=4), viewup="z")
show(t2, at=1)
show(t3, at=2, interactive=1)



"""Threshold the original TetMesh
with a scalar array"""
from vedo import *

tetm = load(dataurl+'limb_ugrid.vtk')
tetm.color('prism').alpha([0,1])

# Threshold the tetrahedral mesh for values in the range:
tetm.threshold(above=0.9, below=1)
tetm.addScalarBar3D(title='chem_0  expression levels', c='k', italic=1)

show([(tetm,__doc__),
       tetm.tomesh(shrink=0.9),
     ], N=2, axes=1)



#
# https://fenicsproject.org/qa/8612/difficulties-with-solving-the-gray-scott-model
from dolfin import *
import numpy as np

# Set parameters
D_u = 8.0e-05
D_v = 4.0e-05
c = 0.022
k = 0.055

dt = 12.0
t_max = 100000

# Form compiler options
parameters["form_compiler"]["optimize"]     = True
parameters["form_compiler"]["cpp_optimize"] = True
parameters["form_compiler"]["representation"] = "uflacs"
set_log_level(30)


# Class representing the initial conditions
class InitialConditions(UserExpression):
    def eval(self, val, x):
        if between(x[0], (1.0, 1.5)) and between(x[1], (1.0, 1.5)):
            val[1] = 0.25*np.power(np.sin(4*np.pi*x[0]), 2)*np.power(np.sin(4*np.pi*x[1]), 2)
            val[0] = 1 - 2*val[1]
        else:
            val[1] = 0
            val[0] = 1
    def value_shape(self):
        return (2,)

# Class for interfacing with the Newton solver
class GrayScottEquations(NonlinearProblem):
    def __init__(self, a, L):
        NonlinearProblem.__init__(self)
        self.L = L
        self.a = a
    def F(self, b, x):
        assemble(self.L, tensor=b)
    def J(self, A, x):
        assemble(self.a, tensor=A)

# Define mesh and function space
p0 = Point(0.0, 0.0)
p1 = Point(2.5, 2.5)
mesh = RectangleMesh(p0, p1, 64, 64)
V = VectorFunctionSpace(mesh, 'CG', 2)

# Define functions
W_init = InitialConditions(degree = 1)
phi    = TestFunction(V)
dp     = TrialFunction(V)
W0     = Function(V)
W      = Function(V)
# Interpolate initial conditions and split functions
W0.interpolate(W_init)
q, p   = split(phi)
u,  v  = split(W)
u0, v0 = split(W0)

# Weak statement of the equations
F1 = u*q*dx -u0*q*dx +D_u*inner(grad(u), grad(q))*dt*dx +u*v*v*q*dt*dx -c*(1-u)*q*dt*dx
F2 = v*p*dx -v0*p*dx +D_v*inner(grad(v), grad(p))*dt*dx -u*v*v*p*dt*dx +(c+k)*v*p*dt*dx
F = F1 + F2

# Compute directional derivative about W in the direction of dp (Jacobian)
a = derivative(F, W, dp)

# Create nonlinear problem and Newton solver
problem = GrayScottEquations(a, F)
solver = NewtonSolver()
#solver.parameters["linear_solver"] = "lu"
#solver.parameters["convergence_criterion"] = "incremental"
solver.parameters["relative_tolerance"] = 1e-3

from vedo.dolfin import *

t = 0.0
W.assign(W0)
while (t < t_max):
    t += dt
    solver.solve(problem, W.vector())
    W0.assign(W)
    u_out, v_out = W.split()
    plot(u_out, Text2D("time = "+str(t)),
         lw=0, warpZfactor=-0.1,
         vmin=0, vmax=1, scalarbar=False, interactive=False)

interactive()




"""
Solution of a particular nonlinear time-dependent
fourth-order equation, known as the Cahn-Hilliard equation.
"""
#https://fenicsproject.org/olddocs/dolfin/1.4.0/python/demo/documented
import random
from dolfin import *
set_log_level(30)


# Class representing the intial conditions
class InitialConditions(UserExpression):
    def __init__(self, **kwargs):
        random.seed(2 + MPI.rank(MPI.comm_world))
        super().__init__(**kwargs)
    def eval(self, values, x):
        values[0] = 0.63 + 0.02 * (0.5 - random.random())
        values[1] = 0.0
    def value_shape(self): return (2,)

# Class for interfacing with the Newton solver
class CahnHilliardEquation(NonlinearProblem):
    def __init__(self, a, L):
        NonlinearProblem.__init__(self)
        self.L = L
        self.a = a
    def F(self, b, x): assemble(self.L, tensor=b)
    def J(self, A, x): assemble(self.a, tensor=A)

# Model parameters
lmbda = 1.0e-02  # surface parameter
dt    = 5.0e-06  # time step
# time stepping family,
# e.g. theta=1 -> backward Euler, theta=0.5 -> Crank-Nicolson
theta = 0.5

# Form compiler options
parameters["form_compiler"]["optimize"] = True
parameters["form_compiler"]["cpp_optimize"] = True

# Create mesh and define function spaces
mesh = UnitSquareMesh(60, 60)
# mesh = UnitSquareMesh.create(60, 60, CellType.Type.triangle)
# V = FunctionSpace(mesh, "Lagrange", 1)
P1 = FiniteElement("Lagrange", mesh.ufl_cell(), 1)
ME = FunctionSpace(mesh, P1 * P1)

# Define trial and test functions
du   = TrialFunction(ME)
q, v = TestFunctions(ME)

# Define functions
u  = Function(ME)  # current solution
u0 = Function(ME)  # solution from previous converged step

# Split mixed functions
dc, dmu = split(du)
c, mu   = split(u)
c0, mu0 = split(u0)

# Create intial conditions and interpolate
u_init = InitialConditions(degree=1)
u.interpolate(u_init)
u0.interpolate(u_init)

# Compute the chemical potential df/dc
c = variable(c)
f = 100 * c ** 2 * (1 - c) ** 2
mu_mid = (1 - theta) * mu0 + theta * mu

# Weak statement of the equations
L0 = c * q - c0 * q + dt * dot(grad(mu_mid), grad(q))
L1 = mu * v - diff(f, c) * v - lmbda * dot(grad(c), grad(v))
L  = (L0 + L1) * dx

# Compute directional derivative about u in the direction of du
a = derivative(L, u, du)

# Create nonlinear problem and Newton solver
problem = CahnHilliardEquation(a, L)
solver = NewtonSolver()
solver.parameters["linear_solver"] = "lu"
solver.parameters["convergence_criterion"] = "incremental"
solver.parameters["relative_tolerance"] = 1e-6

# Step in time
from vedo.dolfin import plot

t = 0
T = 10*dt
scalarbar = False
while t < T:
    t += dt
    u0.vector()[:] = u.vector()
    solver.solve(problem, u.vector())
    if t==T:
        scalarbar = 'horizontal'

    plot(u.split()[0],
         z=t*2e4,
         add=True, # do not clear the canvas
         style=0,
         lw=0,
         scalarbar=scalarbar,
         elevation=-3, # move camera a bit
         azimuth=1,
         text='time: '+str(t*2e4),
         lighting='plastic',
         interactive=0 )

plot()



"""Slice/probe a Volume with a Mesh"""
from vedo import *

vol = load(dataurl+'embryo.slc').mode(1).c('bone')
msh = Paraboloid(res=200).scale(200).pos(100,100,200)
scals = probePoints(vol, msh).getPointArray()
msh.cmap('Spectral', scals).addScalarBar()
show(vol, msh, __doc__, axes=True)



import trimesh
import numpy as np
from vedo import show, settings


# test on a sphere mesh
mesh = trimesh.creation.icosphere()

# create some rays
ray_origins    = np.array([[0, 0, -3], [1,  2, -3]])
ray_directions = np.array([[0, 0,  1], [0, -1,  1]])

# run the mesh-ray query
locations, index_ray, index_tri = mesh.ray.intersects_location(
    ray_origins=ray_origins, ray_directions=ray_directions
)
locs = trimesh.points.PointCloud(locations)

# stack rays into line segments for visualization as Path3D
ray_visualize = trimesh.load_path(
    np.hstack((ray_origins, ray_origins + ray_directions)).reshape(-1, 2, 3)
)

print("The rays hit the mesh at coordinates:\n", locations)
print("The rays with index: {} hit triangles stored at mesh.faces[{}]".format(index_ray, index_tri))

# stack rays into line segments for visualization as Path3D
ray_visualize = trimesh.load_path(
    np.hstack((ray_origins, ray_origins + ray_directions * 5.0)).reshape(-1, 2, 3)
)

# make mesh white-ish
mesh.visual.face_colors = [200, 200, 250, 100]
mesh.visual.face_colors[index_tri] = [255, 0, 0, 255]

show(mesh, ray_visualize, locs, axes=1)



"""Give a cartoony appearance to a 3D mesh"""
from vedo import *

Plotter() # creates a default camera, needed by silhouette()

man = load(dataurl+'man.vtk').lighting('off').c('pink').alpha(0.9)

ted = load(dataurl+'teddy.vtk').lighting('off').c('sienna')
ted.scale(0.4).rotateZ(-45).pos(-1,-1,-1)

show(ted, ted.silhouette(),
     man, man.silhouette(featureAngle=40).lineWidth(3).color('dr'),
     Text2D(__doc__, pos="bottom-center", font="Bongas", s=2, bg='dg'),
     bg='wheat', bg2='lb',
     elevation=-80, zoom=1.2,
    )



"""
This example shows how to animate simultaneously various objects
by specifying event times and durations of the effects
"""
from vedo import *
from vedo.applications import Animation

sp = Sphere(r=0.5).cutWithPlane(origin=(0.15,0,0)).lw(0.1)
cu = Cube().pos(-2,0,0)
tr = Torus().pos(1,0,0).rotateY(80)

vp = Animation()
vp.showProgressBar = True
vp.timeResolution = 0.025  # secs

vp.fadeIn([cu, tr], t=0, duration=0.2)
vp.fadeIn(sp, t=1, duration=2)

vp.move(sp, (2,0,0), style="linear")
vp.rotate(sp, axis="y", angle=180)

vp.fadeOut(sp, t=3, duration=2)
vp.fadeOut(tr, t=4, duration=1)

vp.scale(cu, 0.1, t=5, duration=1)

vp.totalDuration = 4 # can shrink/expand total duration

vp.play()



"""Find the vertices that are connected
to a specific vertex in a mesh
"""
from vedo import *

s = Sphere(c="y", res=12).wireframe()

index = 12 # pick one point
pt = s.points(index)

vtxs = s.connectedVertices(index, returnIds=False)

apt  = Point(pt, c="r", r=15)
cpts = Points(vtxs, c="blue", r=15)

show(s, apt, cpts, __doc__, bg='bb')



# Scatter plot of a gaussian distribution
# with varying color and point sizes
from vedo import *
from vedo.pyplot import plot
import numpy as np

n = 1000
x = np.random.randn(n)
y = np.random.randn(n)

# define what size must have each marker:
marker_sizes = np.sin(2*x)/8

# define a (r,g,b) list of colors for each marker:
marker_cols = np.c_[np.cos(2*x), np.zeros(n), np.zeros(n)]


txt0 = Text2D("A scatter plot of a\n2D gaussian distribution")
plt0 = plot(x, y, ma=0.3, lw=0,        # ma = marker alpha
            marker="*",                # marker style
            xtitle="variable A",
            ytitle="variable B",
           )

txt1 = Text2D("marker size proportional to sin(2x) ")
plt1 = plot(x, y, ma=0.3, lw=0,
            marker="*",                # marker style
            ms=marker_sizes,           # VARIABLE marker sizes
            mc='red',                  # same fixed color for markers
           )

txt2 = Text2D("marker size proportional to sin(2x)\nred level   proportional to cos(2x)")
plt2 = plot(x, y, ma=0.3, lw=0,
            marker=">",                # marker style
            ms=marker_sizes,           # VARIABLE marker sizes
            mc=marker_cols,            # VARIABLE marker colors
           )

show(plt0, txt0, at=0, N=3, size=(1800,500))
show(plt1, txt1, at=1)
show(plt2, txt2, at=2, interactive=True)



"""Build a TetMesh (tetrahedral mesh)
by manually defining vertices and cells"""
from vedo import *

points = [ (0, 0, 0), # first tet
           (1, 0, 0),
           (1, 1, 0),
           (0, 1, 2),
           (3, 3, 3), # second tet
           (4, 3, 3),
           (4, 4, 3),
           (3, 4, 4),
           (2, 5, 3), # third tet
           (3, 5, 3),
           (3, 6, 3),
           (2, 5, 5),
         ]

tets = [[0,1,2,3], [4,5,6,7], [8,9,10,11]]

scal = [10.0, 20.0, 30.0] # cell scalars

# Create the TeTMesh object
tm = TetMesh([points,tets]).addCellArray(scal,'myscal').color('jet')
# tm.color('green') # or set a single color

printc("tetmesh.inputdata():", type(tm.inputdata())) # vtkUnstructuredGrid
printc("points, cells      :", len(tm.points()), len(tm.cells()))

# Optionally convert tm to a Mesh (for visualization)
show([(tm, __doc__),
      (tm.tomesh(),"TetMesh.tomesh()"),
     ],  N=2, axes=1)


"""Cut a mesh with an other mesh
and cap the holes
"""
from vedo import *

p1 = load(dataurl+'motor.byu')

cutmesh = Sphere().y(-0.4).scale(0.4).wireframe().alpha(0.1)

show(p1, cutmesh, __doc__, at=0, N=2, axes=1, viewup="z")

p2 = p1.clone().cutWithMesh(cutmesh)

redcap = p2.cap(returnCap=True).color("r").alpha(0.5)

show(redcap, p2, at=1, interactive=1)



"""
Example of drawing objects on different windows
and/or subwindows within the same window.
We split the main window in many subwindows and draw
somethingon specific windows numbers.
Then open an independent window and draw a shape on it.
"""
print(__doc__)
from vedo import *

##########################################################################
# this is one instance of the class Plotter with 5 raws and 5 columns
vp1 = Plotter(shape=(5,5), axes=0)

# set a different background color for a specific subwindow (the last one)
vp1.renderers[24].SetBackground(0.8, 0.9, 0.9)  # use vtk method SetBackground()

# load the meshes and give them a name
a = vp1.load(dataurl+"shuttle.obj")
b = vp1.load(dataurl+"cessna.vtk").c("red")
c = vp1.load(dataurl+"porsche.ply")

# show a Text2D in each renderer
for i in range(25):
    vp1.show("renderer\nnr."+str(i), at=i)

vp1.show(a, at= 6)
vp1.show(b, at=23)
vp1.show(c, at=24)


##########################################################################
# declare a second independent instance of the class Plotter
# shape can also be given as a string, e.g.:
# shape="2/6" means 2 renderers above and 6 below
# shape="3|1" means 3 renderers on the left and one on the right

s = load(dataurl+'mug.ply')

# Set the position of the horizontal of vertical splitting [0,1]:
#settings.windowSplittingPosition = 0.5

vp2 = Plotter(pos=(500, 250), shape='2/6')

for i in range(len(vp2.renderers)):
    s2 = s.clone(deep=False).color(i)
    vp2.show(s2, 'renderer #'+str(i), at=i)

interactive()



"""Left: build a volume (grey) from a mesh where the
foreground voxels are 1 and the background voxels are 0.

Right: the Volume is isosurfaced.
"""
from vedo import *

s = load(dataurl+"bunny.obj").normalize().wireframe()

v = mesh2Volume(s, spacing=(0.02, 0.02, 0.02)).alpha([0,0.5]).c('blue')

iso = v.isosurface().color("b")

show(v, s.scale(1.05), __doc__, at=0, N=2)

show(iso, at=1, interactive=1)



"""Create a Volume from a numpy array"""
import numpy as np

data_matrix = np.zeros([75, 75, 75], dtype=np.uint8)
# all voxels have value zero except:
data_matrix[0:35,   0:35,  0:35] = 1
data_matrix[35:55, 35:55, 35:55] = 2
data_matrix[55:74, 55:74, 55:74] = 3

from vedo import Volume, show

vol = Volume(data_matrix, c=['white','b','g','r'])
vol.addScalarBar3D()

show(vol, __doc__, axes=1)



"""Click a sphere to highlight it"""
from vedo import Text2D, Sphere, Plotter
import numpy as np

spheres = []
for i in range(25):
    p = np.random.rand(2)
    s = Sphere(r=0.05).pos(p).color('k5')
    s.name = f"sphere nr.{i} at {p}"
    spheres.append(s)

def func(evt):
    if not evt.actor: return
    sil = evt.actor.silhouette().lineWidth(6).c('red5')
    msg.text("You clicked: "+evt.actor.name)
    plt.remove(silcont.pop()).add(sil)
    silcont.append(sil)

silcont = [None]
msg = Text2D("", pos="bottom-center", c='k', bg='r9', alpha=0.8)

plt = Plotter(axes=1, bg='black')
plt.addCallback('mouse click', func)
plt.show(spheres, msg, __doc__, zoom=1.2)



from vedo import *
import numpy as np

pts = np.random.randn(2000,3)*[3,2,4]-[1,2,3]
vpts1 = Points(pts).alpha(0.2).c('blue2')
vpts2 = vpts1.clone().shift(5,6,7).c('green2')

axs = Axes([vpts1, vpts2],  # build axes for this set of objects
           xtitle="X-axis in \mum",
           ytitle="Variable Y in \mum",
           ztitle="Inverted Z in \mum",
           htitle='My \Gamma^2_ijk  plot',
           hTitleFont='Kanopus',
           hTitleJustify='bottom-right',
           hTitleColor='red2',
           hTitleSize=0.035,
           hTitleOffset=(0,0.075,0),
           hTitleRotation=45,
           zHighlightZero=True,
           xyFrameLine=2, yzFrameLine=1, zxFrameLine=1,
           xyFrameColor='red3',
           xyShift=1.05, # move xy 5% above the top of z-range
           yzGrid=True,
           zxGrid=True,
           zxShift=1.0,
           xTitleJustify='bottom-right',
           xTitleOffset=-1.175,
           xLabelOffset=-1.75,
           yLabelRotation=90,
           zInverted=True,
           tipSize=0.25,
          )

show(vpts1, vpts2, axs, "Customizing Axes", viewup='z')



"""A Volume can have multiple
scalars associated to each voxel"""
from vedo import *
import numpy as np

vol = load(dataurl+'vase.vti')
nx, ny, nz = vol.dimensions()
r0,r1 = vol.scalarRange()
vol.addScalarBar3D(title='original voxel scalars')

# create a set of scalars and add it to the Volume
sc1 = np.linspace(r0,r1, num=nx*ny*nz)#.astype(np.uint8)
vol.addPointArray(sc1, "myscalars1")

# create another set of scalars and add it to the Volume
sc2 = np.random.randint(r0,r1, nx*ny*nz)#.astype(np.uint8)
vol.addPointArray(sc2, "myscalars2")

# make SLCImage scalars the active array (can set 0, to pick the first):
printc('Arrays in Volume are:\n', vol.getArrayNames(), invert=1)
vol.getPointArray('SLCImage')

# Build the isosurface of the active scalars,
# but use testscals1 to colorize this isosurface, and then smooth it
iso1 = vol.isosurface().cmap('jet', 'myscalars1').smoothWSinc().lw(0.1)
iso1.addScalarBar3D(title='myscalars1')

iso2 = vol.isosurface().cmap('viridis', 'myscalars2')
iso2.addScalarBar3D(title='myscalars2')

show([(vol, __doc__),
       (iso1,"Colorize isosurface using\nmyscalars1"),
       (iso2,"Colorize isosurface using\nmyscalars2"),
     ], N=3, axes=1)


"""Colorize a large cloud of 1M points by passing
colors and transparencies in the format (R,G,B,A)
"""
from vedo import *
import numpy as np
import time

settings.renderPointsAsSpheres = False
settings.pointSmoothing = False
settings.xtitle = 'red axis'
settings.ytitle = 'green axis'
settings.ztitle = 'blue*alpha axis'

N = 1000000

pts = np.random.rand(N, 3)
RGB = pts * 255
Alpha = pts[:, 2] * 255
RGBA = np.c_[RGB, Alpha]  # concatenate

print("clock starts")
t0 = time.time()

# passing c in format (R,G,B,A) is ~50x faster
pts = Points(pts, r=2, c=RGBA) #fast
#pts = Points(pts, r=2, c=pts, alpha=pts[:, 2]) #slow

t1 = time.time()
print("-> elapsed time:", t1-t0, "seconds for N:", N)

show(pts, __doc__, axes=True)



"""Make a shadow of 2 meshes on the wall"""
from vedo import *

a = load(dataurl+"spider.ply").texture('leather')
a.normalize().rotateZ(-90).addShadow(x=-3, alpha=0.5)

s = Sphere(r=0.3).pos(0.4,0,0.6).addShadow(x=-3)

show(a, s, __doc__, axes=1, viewup="z")



"""Simulation of a block connected 
to a spring in a viscous medium.
"""
from vedo import *

vp = Plotter(interactive=0, axes=0)

L = 0.1  # spring x position at rest
x0 = 0.85  # initial x-coordinate of the block
k = 25  # spring constant
m = 20  # block mass
b = 0.5  # viscosity friction (proportional to velocity)
dt = 0.15  # time step

# initial conditions
v = vector(0, 0, 0.2)
x = vector(x0, 0, 0)
xr = vector(L, 0, 0)
sx0 = vector(-0.8, 0, 0)
offx = vector(0, 0.3, 0)

vp += Box(pos=(0, -0.1, 0), length=2.0, width=0.02, height=0.5)  # surface
vp += Box(pos=(-0.82, 0.15, 0), length=0.04, width=0.50, height=0.3)  # wall

block = Cube(pos=x, side=0.2, c="tomato")
block.addTrail(offset=[0, 0.2, 0], alpha=0.6, lw=2, n=500)
spring = Spring(sx0, x, r=0.06, thickness=0.01)
vp += [block, spring, Text2D(__doc__)]

pb = ProgressBar(0, 300, c="r")
for i in pb.range():
    F = -k * (x - xr) - b * v  # Force and friction
    a = F / m  # acceleration
    v = v + a * dt  # velocity
    x = x + v * dt + 1 / 2 * a * dt ** 2  # position

    block.pos(x)  # update block position and trail
    spring.stretch(sx0, x)  # stretch helix accordingly

    vp.show(elevation=0.1, azimuth=0.1)
    pb.print()

vp.show(interactive=1)



#!/usr/bin/python3
#
"""Interpolate a jpg image to a mesh and plot it"""
from dolfin import *
import matplotlib.pyplot as plt
from vedo.dolfin import plot, download

scale = 0.1
fpath = download("https://vedo.embl.es/examples/data/images/embl_logo.jpg")
img = plt.imread(fpath)
print('Image shape is', img.shape)

img = img[:,:,1]
Nx, Ny = img.shape
mesh = RectangleMesh(Point(0,0,0),
                     Point(Ny*scale, Nx*scale,1), Ny, Nx)

class FE_image(UserExpression):
    def eval_cell(self, value, x, ufc_cell):
        p = Cell(mesh, ufc_cell.index).midpoint()
        i, j = int(p[1]/scale), int(p[0]/scale)
        value[:] = img[-(i+1), j]
    def value_shape(self):
        return ()

y = FE_image()
V = FunctionSpace(mesh, 'Lagrange', 1)
u = Function(V)
u.interpolate(y)

cam = dict(pos=(10.6, 3.71, 22.7),
           focalPoint=(10.6, 3.71, -1.04e-3),
           viewup=(0, 1.00, 0),
           distance=22.7,
           clippingRange=(21.3, 24.6)) # press C to get this lines of code

plot(u,
     text=__doc__,
     camera=cam,
     lw=0.1,
     cmap='Greens_r',
     size=(600,300)
     )


"""Cut a mesh with another mesh"""
from vedo import *

embryo = load(dataurl+"embryo.tif").isosurface(30).normalize()
txt = Text2D(__doc__, c='b', bg='lb')

# mesh used to cut:
msh = Ellipsoid().scale(0.4).pos(2.8, 1.5, 1.5).wireframe()

# make a working copy and cut it with the ellipsoid
cutembryo = embryo.clone().cutWithMesh(msh).c("gold").bc("t")

show(embryo, msh,    at=0, N=2, axes=1, viewup="z")
show(cutembryo, txt, at=1, interactive=1)



"""Self organizing maps"""
# -----------------------------------------------------------------------------
# Copyright 2019 (C) Nicolas P. Rougier
# Released under a BSD two-clauses license
#
# References: Kohonen, Teuvo. Self-Organization and Associative Memory.
#             Springer, Berlin, 1984.
# https://github.com/rougier/ML-Recipes/blob/master/recipes/ANN/som.py
# -----------------------------------------------------------------------------
import numpy as np
import scipy.spatial
from vedo import *


class SOM:

    def __init__(self, shape, distance):
        self.codebook = np.random.uniform(0, 1, shape)
        self.distance = distance / distance.max()
        self.samples  = []

    def learn(self, n_epoch=10000, sigma=(0.25,0.01), lrate=(0.5,0.01)):

        t = np.linspace(0, 1, n_epoch)
        lrate = lrate[0] * (lrate[1] / lrate[0]) ** t
        sigma = sigma[0] * (sigma[1] / sigma[0]) ** t
        I = np.random.randint(0, len(self.samples), n_epoch)
        self.samples = self.samples[I]
        pts = Points(self.samples, r=2, c='darkred')
        doc = Text2D(__doc__)

        pb = ProgressBar(0,n_epoch)
        for i in pb.range():
            pb.print("epochs")
            # Get random sample
            data = self.samples[i]

            # Get index of nearest node (minimum distance)
            winner = np.argmin(((self.codebook - data)**2).sum(axis=-1))

            # Gaussian centered on winner
            G = np.exp(-self.distance[winner]**2 / sigma[i]**2)

            # Move nodes towards sample according to Gaussian
            self.codebook -= lrate[i] * G[..., np.newaxis] * (self.codebook-data)

            # Draw network
            if i>500 and not i%20 or i==n_epoch-1:
                x, y, z = [self.codebook[:,i].reshape(n,n) for i in range(3)]
                grd = Grid(resx=n-1, resy=n-1, c='g')
                grd.wireframe(False).lw(0.5).bc('lightblue').flat()
                grdpts = grd.points()
                for i in range(n):
                    for j in range(n):
                        grdpts[i*n+j] = (x[i,j], y[i,j], z[i,j])
                grd.points(grdpts)
                show(doc, pts, grd, axes=6, azimuth=2, interactive=False)

        interactive()
        return [self.codebook[:,i].reshape(n,n) for i in range(3)]

# -------------------------------------------------------------------------------
if __name__ == "__main__":

    n = 25
    X, Y = np.meshgrid(np.linspace(0, 1, n), np.linspace(0, 1, n))
    P = np.c_[X.ravel(), Y.ravel()]
    D = scipy.spatial.distance.cdist(P, P)

    s = Sphere(res=90).cutWithPlane(origin=(0,-.3,0), normal='y').clean(0.01)

    som = SOM((len(P), 3), D)
    som.samples = s.points()
    som.learn(n_epoch=7000, sigma=(1, 0.01), lrate=(1, 0.01))



"""Remove points and cells from a mesh
which are closest to a specified point.
"""
from vedo import *

pu = load(dataurl+'apple.ply')
pu.c('lightgreen').bc('tomato').lw(0.1)

pt = [1, 0.5, 1]
ids = pu.closestPoint(pt, N=200, returnPointId=True)

pu.deletePoints(ids, renamePoints=1)

show(Point(pt), pu, __doc__, axes=1)



"""Customizing axes style
(40+ control parameters!)
Title font: """
from vedo import Box, Lines, Points, Spline, show, settings

settings.defaultFont = 'Theemim'

# an invisible box:
world = Box(pos=(2.7,0,0), size=(12,10,8), alpha=0)

# a dummy spline with its shadow on the xy plane
pts = Points([(-2,-3.2,-1.5), (3,-1.2,-2), (7,3,4)], r=12)
spl = Spline(pts, res=50).addShadow(z=-4) # make spline and add its shadow at z=-4
lns = Lines(spl, spl.shadow)              # join spline points with its own shadow

# make a dictionary of axes options
axes_opts = dict(
    xtitle='My variable \Omega^\lowerxi_lm  in units of \mum^3', # latex-style syntax
    ytitle='This is my highly\ncustomized y-axis',
    ztitle='z in units of Å', # many unicode chars are supported (type: vedo -r fonts)
    yValuesAndLabels=[(-3.2,'Mark^a_-3.2'), (-1.2,'Carmen^b_-1.2'), (3,'John^c_3')],
    textScale=1.3,       # make all text 30% bigger
    numberOfDivisions=5, # approximate number of divisions on longest axis
    axesLineWidth= 2,
    gridLineWidth= 1,
    zxGrid2=True,        # show zx plane on opposite side of the bounding box
    yzGrid2=True,        # show yz plane on opposite side of the bounding box
    xyPlaneColor='green7',
    xyGridColor='dg',    # darkgreen line color
    xyAlpha=0.2,         # grid opacity
    xTitlePosition=0.5,  # title fractional positions along axis
    xTitleJustify="top-center", # align title wrt to its axis
    yTitleSize=0.02,
    yTitleBox=True,
    yTitleOffset=0.05,
    yLabelOffset=0.4,
    yHighlightZero=True, # draw a line highlighting zero position if in range
    yHighlightZeroColor='red',
    zLineColor='blue',
    zTitleColor='blue',
    zTitleBackfaceColor='v', # violet color of axis title backface
    labelFont="Quikhand",
    yLabelSize=0.025,    # size of the numeric labels along Y axis
    yLabelColor='dg',    # color of the numeric labels along Y axis
)

show(world, pts, spl, lns, __doc__+settings.defaultFont, axes=axes_opts)




from vedo import *

# mode = 1 is maximum projection (default is 0=composite)
v1 = load(dataurl+'embryo.tif').mode(1)
t1 = Text2D('Original volume', c='lg')

# cutoff range is roughly in the range of 1 / size of object
v2 = v1.clone().frequencyPassFilter(highcutoff=.001, order=1).mode(1)
t2 = Text2D('High freqs in the FFT\nare cut off:', c='lb')

printHistogram(v1, logscale=1, horizontal=1, c='g')
printHistogram(v2, logscale=1, horizontal=1, c='b')
v1.addScalarBar3D(c='w')
v2.addScalarBar3D(c='w')

show([(v1,t1), (v2,t2)], N=2, bg='bb', zoom=1.5, axes=dict(digits=2))
#write(v2, 'embryo_filtered.vti')



"""
Visualize a Fenics/dolfin mesh.
Select mesh and press X to slice it.
"""
import dolfin
from vedo.dolfin import plot, download

fpath = download("https://vedo.embl.es/examples/data/dolfin_fine.xml")
mesh1 = dolfin.Mesh(fpath)

plot(mesh1)

# show another light-green mesh in a new plotter window,
# show file header too as an additional text comment
mesh2 = dolfin.UnitCubeMesh(8,8,8)

plot(mesh2, text=__doc__, color='lg', new=True)



from vedo import dataurl, Volume
from vedo.applications import IsosurfaceBrowser

vol = Volume(dataurl+'head.vti')

plt = IsosurfaceBrowser(vol, c='gold') # Plotter instance

plt.show(axes=7, bg2='lb')



"""Generate 3 random sets of points and align
them using Procrustes method."""
from __future__ import division, print_function
from random import uniform as u

from vedo import Plotter, procrustesAlignment, Points

vp = Plotter(shape=[1, 2], axes=2, sharecam=0)

N = 15  # number of points
x = 1.0  # add some randomness

pts1 = [(u(0, x), u(0, x), u(0, x) + i) for i in range(N)]
pts2 = [(u(0, x) + 3, u(0, x) + i / 2 + 2, u(0, x) + i + 1) for i in range(N)]
pts3 = [(u(0, x) + 4, u(0, x) + i / 4 - 3, u(0, x) + i - 2) for i in range(N)]

vpts1 = Points(pts1, c="r").legend("set1")
vpts2 = Points(pts2, c="g").legend("set2")
vpts3 = Points(pts3, c="b").legend("set3")

vp.show(vpts1, vpts2, vpts3, __doc__, at=0)

# find best alignment among the n sets of Points,
# return an Assembly object formed by the aligned sets
aligned = procrustesAlignment([vpts1, vpts2, vpts3])

# print(aligned.info['transform'])

vp.show(aligned, at=1, interactive=1)



"""Setting illumination properties:
ambient, diffuse
specular, specularPower, specularColor.
"""
#https://lorensen.github.io/VTKExamples/site/Python/Rendering/SpecularSpheres
from vedo import Plotter, Arrow, Light, dataurl


vp = Plotter(axes=1)

ambient, diffuse, specular = 0.1, 0., 0.
specularPower, specularColor= 20, 'white'

for i in range(8):
    s = vp.load(dataurl+'apple.ply').c('gold')
    s.normalize().pos((i%4)*2.2, int(i<4)*3, 0)

    #s.phong()
    s.flat()

    # modify the default with specific values
    s.lighting('default', ambient, diffuse, specular, specularPower, specularColor)
    #ambient += 0.125
    diffuse += 0.125
    specular += 0.125

vp += __doc__
vp.show()

print('Adding a light source..')
p = (3, 1.5, 3)
f = (3, 1.5, 0)
vp += [Arrow(p,f, s=0.01, c='gray', alpha=0.2), Light(pos=p, focalPoint=f)]
vp.show()



"""
Extract submesh boundaries.
"""
# https://fenicsproject.discourse.group/t/output-parts-of-boundary/537
from fenics import *
from mshr import *
from numpy import array
from numpy.linalg import norm

domain = Box(Point(0,0,0), Point(10,10,10)) - Sphere(Point(5,5,5), 3)
mesh = generate_mesh(domain, 32)
exterior = BoundaryMesh(mesh, "exterior")


def inSphere(x):
    v = x - array([5, 5, 5])
    return norm(v) < 3 + 1e2 * DOLFIN_EPS

class SphereDomain(SubDomain):
    def inside(self, x, on_boundary):
        return inSphere(x)

class BoxDomain(SubDomain):
    def inside(self, x, on_boundary):
        return not inSphere(x)

sph = SubMesh(exterior, SphereDomain())
box = SubMesh(exterior, BoxDomain())


from vedo.dolfin import plot

plot(sph, at=0, N=2, c='red', text=__doc__)
plot(box, at=1, wireframe=True)



"""Triangulate arbitrary line contours in 2D.
The contours may be concave, and even contain holes.
"""
from vedo import *

# let's create two bidimentional contour lines
s1 = Star(line=True, n=9)
s2 = Star(line=True, n=5, r1=0.3, r2=0.4).x(0.12)
# merge the 2 lines and triangulate the inner region
sm = merge(s1, s2).triangulate().c('lightsalmon').lw(0.1)

show([(s1,s2,__doc__), sm], N=2, axes=1)



"""Expand and reconstruct any surface
(here a simple box) into spherical harmonics"""
# Expand an arbitrary closed shape in spherical harmonics
# using SHTOOLS (https://shtools.oca.eu/shtools/)
# and then truncate the expansion to a specific lmax and
# reconstruct the projected points in red
from vedo import *
import numpy as np
from scipy.interpolate import griddata
import pyshtools

###########################################################################
lmax = 8              # maximum degree of the spherical harm. expansion
N    = 50             # number of grid intervals on the unit sphere
rmax = 500            # line length
x0 = [250, 250, 250]  # set SPH sphere at this position
###########################################################################

x0 = np.array(x0)
surface = Box(pos=x0+[10,20,30], size=(300,150,100)).color('grey').alpha(0.2)

############################################################
# cast rays from the sphere center and find intersections
agrid, pts = [], []
for th in np.linspace(0, np.pi, N, endpoint=True):
    longs = []
    for ph in np.linspace(0, 2*np.pi, N, endpoint=False):
        p = spher2cart(rmax, th, ph)
        intersections = surface.intersectWithLine(x0, x0+p)
        if len(intersections):
            value = mag(intersections[0]-x0)
            longs.append(value)
            pts.append(intersections[0])
        else:
            printc('No hit for theta, phi =', th, ph, c='r')
            longs.append(rmax)
            pts.append(p)
    agrid.append(longs)
agrid = np.array(agrid)

hits = Points(pts).cmap('jet', agrid.ravel()).addScalarBar3D(title='scalar distance to x_0')
show([surface, hits, Point(x0), __doc__], at=0, N=2, axes=1)

#############################################################
grid = pyshtools.SHGrid.from_array(agrid)
clm = grid.expand()
grid_reco = clm.expand(lmax=lmax).to_array()  # cut "high frequency" components

#############################################################
# interpolate to a finer grid
ll = []
for i, long in enumerate(np.linspace(0, 360, num=grid_reco.shape[1], endpoint=False)):
    for j, lat in enumerate(np.linspace(90, -90, num=grid_reco.shape[0], endpoint=True)):
        th = np.deg2rad(90 - lat)
        ph = np.deg2rad(long)
        p = spher2cart(grid_reco[j][i], th, ph)
        ll.append((lat, long))

radii = grid_reco.T.ravel()
n = 200j
lnmin, lnmax = np.array(ll).min(axis=0), np.array(ll).max(axis=0)
grid = np.mgrid[lnmax[0]:lnmin[0]:n, lnmin[1]:lnmax[1]:n]
grid_x, grid_y = grid
grid_reco_finer = griddata(ll, radii, (grid_x, grid_y), method='cubic')

pts2 = []
for i, long in enumerate(np.linspace(0, 360, num=grid_reco_finer.shape[1], endpoint=False)):
    for j, lat in enumerate(np.linspace(90, -90, num=grid_reco_finer.shape[0], endpoint=True)):
        th = np.deg2rad(90 - lat)
        ph = np.deg2rad(long)
        p = spher2cart(grid_reco_finer[j][i], th, ph)
        pts2.append(p+x0)

show(Points(pts2, c="r", alpha=0.5), surface,
     'Spherical harmonics\nexpansion of order '+str(lmax),
     at=1, interactive=True)



"""
FEniCS tutorial demo program: Diffusion of a Gaussian hill.

  u'= Laplace(u) + f  in a square domain
  u = u_D             on the boundary
  u = u_0             at t = 0

  u_D = f = 0

The initial condition u_0 is chosen as a Gaussian hill.
"""
# https://fenicsproject.org/pub/tutorial/html/._ftut1006.html
from __future__ import print_function
from fenics import *
set_log_level(30)


num_steps = 50  # number of time steps
dt = 0.02       # time step size

# Create mesh and define function space
nx = ny = 30
mesh = RectangleMesh(Point(-2, -2), Point(2, 2), nx, ny)
V = FunctionSpace(mesh, 'P', 1)

# Define boundary condition
def boundary(x, on_boundary):
    return on_boundary

bc = DirichletBC(V, Constant(0), boundary)

# Define initial value
u_0 = Expression('exp(-5*pow(x[0],2) - 5*pow(x[1],2))', degree=2)
u_n = interpolate(u_0, V)

# Define variational problem
u = TrialFunction(V)
v = TestFunction(V)
f = Constant(0)

F = u*v*dx + dt*dot(grad(u), grad(v))*dx - (u_n + dt*f)*v*dx
a, L = lhs(F), rhs(F)

############################################################# vedo
from vedo.dolfin import plot, Latex

f = r'\frac{\partial u}{\partial t}=\nabla^2 u+f~\mathrm{in}~\Omega\times(0,T]'
formula = Latex(f, pos=(-.4,-.8, .1), s=0.6, c='w')
formula.crop(0.2, 0.4) # crop top and bottom 20% and 40%

# Time-stepping
u = Function(V)
for n in range(num_steps):

    # Compute solution
    solve(a == L, u, bc)

    # Plot solution
    plot(u, formula, scalarbar=False, interactive=False)

    # Update previous solution
    u_n.assign(u)

plot()



from vedo import *
from vedo.pyplot import histogram
import numpy as np

N = 2000
x = np.random.randn(N) * 1.0
y = np.random.randn(N) * 1.5

# hexagonal binned histogram:
histo = histogram(x, y,
                  bins=10,
                  mode='hexbin',
                  xtitle="\sigma_x =1.0",
                  ytitle="\sigma_y =1.5",
                  ztitle="counts",
                  fill=True,
                  cmap='terrain',
                  )
# add a formula:
f = r'f(x, y)=A \exp \left(-\left(\frac{\left(x-x_{o}\right)^{2}}'
f+= r'{2 \sigma_{x}^{2}}+\frac{\left(y-y_{o}\right)^{2}}'
f+= r'{2 \sigma_{y}^{2}}\right)\right)'
formula = Latex(f, c='k', s=1.5).rotateX(90).rotateZ(90).pos(1.5,-2,1)

show(histo, formula, axes=1, viewup='z')



"""Generate a scalar field by the signed distance from a mesh,
optionally save it to a vti file,
then extract an isosurface from the 3d image.
"""
from vedo import *

mesh = load(dataurl+"apple.ply").subdivide()

# Generate signed distance volume
vol = volumeFromMesh(mesh,
                     dims=(40,40,40),
                     bounds=(-1.3, 1.3, -1.3, 1.3, -1.3, 1.3),
                     signed=True,
                     negate=True, # invert sign
)
#write(vol, 'stack.vti')

iso = vol.isosurface(threshold=-0.01)

pts = Points(mesh.points())

show(iso, pts, __doc__, axes=1)



"""Generate the silhouette of a mesh
as seen along a specified direction

Axes font: """
# Source: Zhi-Qiang Zhou (https://github.com/zhouzq-thu)
from vedo import *

settings.defaultFont = 'Kanopus'

plt = Plotter(title="Example of projectOnPlane()")

s = Hyperboloid().rotateX(20)
pts = s.points()
n = len(pts)

plt += [s, __doc__+settings.defaultFont]

# orthogonal projection ###############################
plane1 = Plane(pos=(2,0,2), normal=(1,0,1), sx=5).alpha(0.1)
so = s.clone().projectOnPlane(plane1).c('y')
plt += [plane1, so, so.silhouette('2d')]
pts1 = so.silhouette('2d').points()

# perspective projection ##############################
plane2 = Plane(pos=(3,3,3), normal=(1,1,1), sx=5).alpha(0.1)
point = [6, 6, 6]
sp = s.clone().projectOnPlane(plane2, point=point).c('m')
plt += [plane2, sp, sp.silhouette('2d')]

# oblique projection ##################################
plane3 = Plane(pos=(4,8,-4), normal=(-1,0,1), sx=5).alpha(0.1)
sob = s.clone().projectOnPlane(plane3, direction=(1,2,-1)).c('g')
plt += [plane3, sob, sob.silhouette('2d')]
pts2 = sob.silhouette('2d').points()

# draw the lines
for i in range(0, n, int(n/20)):
     plt += Line(pts1[i], pts[i], c='k', alpha=0.2)
     plt += Line(point,   pts[i], c='k', alpha=0.2)
     plt += Line(pts2[i], pts[i], c='k', alpha=0.2)

plt.show(axes=dict(xtitle='X-axis in \mum',
                   ytitle='Y-axis in \mum',
                   ztitle='Z-axis in \mum',
                   yzGrid=False,
                   textScale=1.5,
                  ),
         )



"""Heat equation in moving media."""
# Credits: Jan Blechta
# https://github.com/blechta/fenics-handson/blob/master/heatconv
from dolfin import *
set_log_level(30)


# Create mesh and build function space
mesh = UnitSquareMesh(30, 30, "crossed")
V = FunctionSpace(mesh, "Lagrange", 1)

# Create boundary markers
tdim = mesh.topology().dim()
boundary_parts = MeshFunction("size_t", mesh, tdim - 1)
left = AutoSubDomain(lambda x: near(x[0], 0.0))
right = AutoSubDomain(lambda x: near(x[0], 1.0))
bottom = AutoSubDomain(lambda x: near(x[1], 0.0))
left.mark(boundary_parts, 1)
right.mark(boundary_parts, 2)
bottom.mark(boundary_parts, 2)

# Initial condition and right-hand side
ic = Expression("""pow(x[0] - 0.25, 2) + pow(x[1] - 0.25, 2) < 0.2*0.2
                   ? -25.0 * ((pow(x[0] - 0.25, 2) + pow(x[1] - 0.25, 2)) - 0.2*0.2)
                   : 0.0""", degree=1,
)
f = Expression("""pow(x[0] - 0.75, 2) + pow(x[1] - 0.75, 2) < 0.2*0.2
                  ? 1.0
                  : 0.0""", degree=1,
)

# Equation coefficients
K = Constant(1e-2)  # thermal conductivity
g = Constant(0.01)  # Neumann heat flux
b = Expression(("-(x[1] - 0.5)", "x[0] - 0.5"), degree=1)  # convecting velocity

# Define boundary measure on Neumann part of boundary
dsN = Measure("ds", subdomain_id=1, subdomain_data=boundary_parts)

# Define steady part of the equation
def operator(u, v):
    return (K * inner(grad(u), grad(v)) - f * v + dot(b, grad(u)) * v
    ) * dx - K * g * v * dsN

# Define trial and test function and solution at previous time-step
u = TrialFunction(V)
v = TestFunction(V)
u0 = Function(V)

# Time-stepping parameters
dt = 0.02
theta = Constant(0.5)  # Crank-Nicolson scheme

# Define time discretized equation
F = ((1.0 / dt) * inner(u - u0, v) * dx
    + theta * operator(u, v)
    + (1.0 - theta) * operator(u0, v)
)

# Define boundary condition
bc = DirichletBC(V, Constant(0.0), boundary_parts, 2)

# Prepare solution function and solver
u = Function(V)
problem = LinearVariationalProblem(lhs(F), rhs(F), u, bc)
solver = LinearVariationalSolver(problem)

# Prepare initial condition
u0.interpolate(ic)
u.interpolate(ic)


######################################################Time-stepping
from vedo.dolfin import *

t = 0.0
while t < 3:
    solver.solve()

    plot(u,
        text=__doc__+"\nTemperature at t = %g" % t,
        style=2,
        axes=3,
        lw=0, # no mesh edge lines
        warpZfactor=0.1,
        isolines={"n": 12, "lw":1, "c":'black', "alpha":0.1},
        scalarbar=False,
        interactive=False,
    )

    # Move to next time step
    u0.assign(u)
    t += dt

interactive()



"""Create a simple Play/Pause app with a timer event
You can interact with the scene during the loop!
..press q to quit"""
import vedo


def buttonfunc():
    global timerId
    plotter.timerCallback("destroy", timerId)
    if "Play" in button.status():
        # instruct to call handle_timer() every 10 msec:
        timerId = plotter.timerCallback("create", dt=10)
    button.switch()

def handle_timer(event):
    ### Animate your stuff here ######################################
    earth.rotateZ(1)            # rotate the Earth by 1 deg
    plotter.render()


plotter = vedo.Plotter()

timerId = None
button = plotter.addButton(buttonfunc, states=[" Play ","Pause"], size=40)
evntId = plotter.addCallback("timer", handle_timer)

earth = vedo.Earth()

plotter.show(earth, __doc__, axes=1, bg2='b9', viewup='z')



"""Dijkstra algorithm to compute the graph geodesic.

Takes as input a polygonal mesh and performs
a shortest path calculation 20 times.
"""
from vedo import *

s = Sphere(r=1.02, res=200).clean(0.007).wireframe().alpha(0.02)

paths = []
for i in range(20):
    paths.append(s.geodesic(2500, i*700))
    # print(paths[-1].info['CumulativeWeights'])

show(s, Earth(), __doc__, paths, bg2='lb', viewup="z")



import vtk
from vtk.util.numpy_support import numpy_to_vtk
import vedo.docs as docs
import vedo.utils as utils
from vedo.base import BaseGrid
from vedo.mesh import Mesh
from vedo.colors import printc
import numpy as np

__doc__ = (
    """
Support for tetrahedral meshes.
"""
    + docs._defs
)

__all__ = ["TetMesh", "delaunay3D", "tetralize"]


##########################################################################
def delaunay3D(mesh, alphaPar=0, tol=None, boundary=False):
    """Create 3D Delaunay triangulation of input points."""
    deln = vtk.vtkDelaunay3D()
    if utils.isSequence(mesh):
        pd = vtk.vtkPolyData()
        vpts = vtk.vtkPoints()
        vpts.SetData(numpy_to_vtk(np.ascontiguousarray(mesh), deep=True))
        pd.SetPoints(vpts)
        deln.SetInputData(pd)
    else:
        deln.SetInputData(mesh.GetMapper().GetInput())
    deln.SetAlpha(alphaPar)
    deln.AlphaTetsOn()
    deln.AlphaTrisOff()
    deln.AlphaLinesOff()
    deln.AlphaVertsOff()
    if tol:
        deln.SetTolerance(tol)
    deln.SetBoundingTriangulation(boundary)
    deln.Update()
    m = TetMesh(deln.GetOutput())
    return m


def tetralize(dataset, tetsOnly=True):
    """Tetralize any type of dataset.
    If tetsOnly is True will cull all 1D and 2D cells from the output.

    Return a TetMesh.

    Example:

        .. code-block:: python

            from vedo import *
            ug = loadUnStructuredGrid(datadir+'ugrid.vtk')
            tmesh = tetralize(ug)
            tmesh.write('ugrid.vtu').show(axes=1)
    """
    tt = vtk.vtkDataSetTriangleFilter()
    tt.SetInputData(dataset)
    tt.SetTetrahedraOnly(tetsOnly)
    tt.Update()
    m = TetMesh(tt.GetOutput())
    return m


##########################################################################
class TetMesh(vtk.vtkVolume, BaseGrid):
    """The class describing tetrahedral meshes."""

    def __init__(self, inputobj=None,
                 c=('r','y','lg','lb','b'), #('b','lb','lg','y','r')
                 alpha=(0.5, 1),
                 alphaUnit=1,
                 mapper='tetra',
                 ):

        BaseGrid.__init__(self)

        self.useArray = 0

        inputtype = str(type(inputobj))
        #printc('TetMesh inputtype', inputtype)

        ###################
        if inputobj is None:
            self._data = vtk.vtkUnstructuredGrid()

        elif isinstance(inputobj, vtk.vtkUnstructuredGrid):
            self._data = inputobj

        elif isinstance(inputobj, vtk.vtkRectilinearGrid):
            r2t = vtk.vtkRectilinearGridToTetrahedra()
            r2t.SetInputData(inputobj)
            r2t.RememberVoxelIdOn()
            r2t.SetTetraPerCellTo6()
            r2t.Update()
            self._data = r2t.GetOutput()

        elif isinstance(inputobj, vtk.vtkDataSet):
            r2t = vtk.vtkDataSetTriangleFilter()
            r2t.SetInputData(inputobj)
            #r2t.TetrahedraOnlyOn()
            r2t.Update()
            self._data = r2t.GetOutput()

        elif isinstance(inputobj, str):
            from vedo.io import download, loadUnStructuredGrid
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            ug = loadUnStructuredGrid(inputobj)
            tt = vtk.vtkDataSetTriangleFilter()
            tt.SetInputData(ug)
            tt.SetTetrahedraOnly(True)
            tt.Update()
            self._data = tt.GetOutput()

        elif utils.isSequence(inputobj):
            # if "ndarray" not in inputtype:
            #     inputobj = np.array(inputobj)
            self._data = self._buildtetugrid(inputobj[0], inputobj[1])

        ###################
        if 'tetra' in mapper:
            self._mapper = vtk.vtkProjectedTetrahedraMapper()
        elif 'ray' in mapper:
            self._mapper = vtk.vtkUnstructuredGridVolumeRayCastMapper()
        elif 'zs' in mapper:
            self._mapper = vtk.vtkUnstructuredGridVolumeZSweepMapper()
        elif isinstance(mapper, vtk.vtkMapper):
            self._mapper = mapper
        else:
            printc('Unknown mapper type', [mapper], c='r')
            raise RuntimeError()

        self._mapper.SetInputData(self._data)
        self.SetMapper(self._mapper)
        self.color(c).alpha(alpha)
        if alphaUnit:
            self.GetProperty().SetScalarOpacityUnitDistance(alphaUnit)

        # remember stuff:
        self._color = c
        self._alpha = alpha
        self._alphaUnit = alphaUnit
        #-----------------------------------------------------------

    def _update(self, data):
        self._data = data
        self._mapper.SetInputData(data)
        self._mapper.Modified()
        return self


    def _buildtetugrid(self, points, cells):
        ug = vtk.vtkUnstructuredGrid()

        if len(points) == 0:
            return ug
        if not utils.isSequence(points[0]):
            return ug

        if len(cells) == 0:
            return ug

        if not utils.isSequence(cells[0]):
            tets=[]
            nf=cells[0]+1
            for i, cl in enumerate(cells):
                if i==nf or i==0:
                    k = i+1
                    nf = cl+k
                    cell = [cells[j+k] for j in range(cl)]
                    tets.append(cell)
            cells = tets

        sourcePoints = vtk.vtkPoints()
        varr = numpy_to_vtk(np.ascontiguousarray(points), deep=True)
        sourcePoints.SetData(varr)
        ug.SetPoints(sourcePoints)

        sourceTets = vtk.vtkCellArray()
        for f in cells:
            ele = vtk.vtkTetra()
            pid = ele.GetPointIds()
            for i, fi in enumerate(f):
                pid.SetId(i, fi)
            sourceTets.InsertNextCell(ele)
        ug.SetCells(vtk.VTK_TETRA, sourceTets)
        return ug


    def clone(self):
        """Clone the ``TetMesh`` object to yield an exact copy."""
        ugCopy = vtk.vtkUnstructuredGrid()
        ugCopy.DeepCopy(self._data)

        cloned = TetMesh(ugCopy)
        pr = vtk.vtkVolumeProperty()
        pr.DeepCopy(self.GetProperty())
        cloned.SetProperty(pr)

        #assign the same transformation to the copy
        cloned.SetOrigin(self.GetOrigin())
        cloned.SetScale(self.GetScale())
        cloned.SetOrientation(self.GetOrientation())
        cloned.SetPosition(self.GetPosition())

        cloned._mapper.SetScalarMode(self._mapper.GetScalarMode())
        cloned.name = self.name
        return cloned


    def threshold(self, name=None, above=None, below=None):
        """
        Threshold the tetrahedral mesh by a cell scalar value.
        Reduce to only tets which satisfy the threshold limits.
        """
        th = vtk.vtkThreshold()
        th.SetInputData(self._data)
        ns = self.getArrayNames()

        if name is None:
            if len(ns['CellData']):
                name=ns['CellData'][0]
                th.SetInputArrayToProcess(0,0,0, 1, name)
            elif len(ns['PointData']):
                name=ns['PointData'][0]
                th.SetInputArrayToProcess(0,0,0, 0, name)
            if name is None:
                printc("threshold(): Cannot find active array. Skip.", c='r')
                return self
        else:
            if self.useCells:
                th.SetInputArrayToProcess(0,0,0, 1, name)
            else:
                th.SetInputArrayToProcess(0,0,0, 0, name)

        if above is not None and below is not None:
            if above<below:
                th.ThresholdBetween(above, below)
            elif above==below:
                return self

        elif above is not None:
            th.ThresholdByUpper(above)

        elif below is not None:
            th.ThresholdByLower(below)

        th.Update()
        ugrid = th.GetOutput()
        return self._update(ugrid)



    def decimate(self, scalarsName, fraction=0.5, N=None):
        """
        Downsample the number of tets in a TetMesh to a specified fraction.

        :param float fraction: the desired final fraction of the total.
        :param int N: the desired number of final tets

        .. note:: Setting ``fraction=0.1`` leaves 10% of the original nr of tets.
        """
        decimate = vtk.vtkUnstructuredGridQuadricDecimation()
        decimate.SetInputData(self._data)
        decimate.SetScalarsName(scalarsName)

        if N:  # N = desired number of points
            decimate.SetNumberOfTetsOutput(N)
        else:
            decimate.SetTargetReduction(1-fraction)
        decimate.Update()
        return self._update(decimate.GetOutput())


    def subdvide(self):
        """Increase the number of tets of a TetMesh.
        Subdivide one tetrahedron into twelve for every tetra."""
        sd = vtk.vtkSubdivideTetra()
        sd.SetInputData(self._data)
        sd.Update()
        return self._update(sd.GetOutput())


    def isosurface(self, threshold=True):
        """Return a ``Mesh`` isosurface.

        :param float,list threshold: value or list of values to draw the isosurface(s)
        """
        if not self._data.GetPointData().GetScalars():
            self.mapCellsToPoints()
        scrange = self._data.GetPointData().GetScalars().GetRange()
        cf = vtk.vtkContourFilter() #vtk.vtkContourGrid()
        cf.SetInputData(self._data)

        if utils.isSequence(threshold):
            cf.SetNumberOfContours(len(threshold))
            for i, t in enumerate(threshold):
                cf.SetValue(i, t)
            cf.Update()
        else:
            if threshold is True:
                threshold = (2 * scrange[0] + scrange[1]) / 3.0
                #print('automatic threshold set to ' + utils.precision(threshold, 3), end=' ')
                #print('in [' + utils.precision(scrange[0], 3) + ', ' + utils.precision(scrange[1], 3)+']')
            cf.SetValue(0, threshold)
            cf.Update()

        clp = vtk.vtkCleanPolyData()
        clp.SetInputData(cf.GetOutput())
        clp.Update()
        msh = Mesh(clp.GetOutput(), c=None).phong()
        msh._mapper.SetLookupTable(utils.ctf2lut(self))
        return msh


    def slice(self, origin=(0,0,0), normal=(1,0,0)):
        """Return a 2D slice of the mesh by a plane passing through origin and
        assigned normal."""
        strn = str(normal)
        if strn   ==  "x": normal = (1, 0, 0)
        elif strn ==  "y": normal = (0, 1, 0)
        elif strn ==  "z": normal = (0, 0, 1)
        elif strn == "-x": normal = (-1, 0, 0)
        elif strn == "-y": normal = (0, -1, 0)
        elif strn == "-z": normal = (0, 0, -1)
        plane = vtk.vtkPlane()
        plane.SetOrigin(origin)
        plane.SetNormal(normal)

        cc = vtk.vtkCutter()
        cc.SetInputData(self._data)
        cc.SetCutFunction(plane)
        cc.Update()
        msh = Mesh(cc.GetOutput()).flat().lighting('ambient')
        msh._mapper.SetLookupTable(utils.ctf2lut(self))
        return msh




























"""Histogram of 2 variables"""
from vedo import *
from vedo.pyplot import histogram
import numpy as np

n = 10000
x = np.random.normal(2, 1, n)*2 + 3
y = np.random.normal(1, 1, n)*1 + 7
xm, ym = np.mean(x), np.mean(y)

h = histogram(x, y,
              bins=50,
              aspect=4/3,
              cmap='Blues',
              title='2D Gauss histo',
              scalarbar=True,
             )

# add some object to the plot
h += Marker('*', s=0.3, c='r').pos(xm, ym, 0.1)

show(h)



"""Erode or dilate a Volume
by replacing a voxel with the max/min
over an ellipsoidal neighborhood"""
from vedo import *

em = Volume(dataurl+'embryo.tif').lighting('plastic').printHistogram(logscale=1)

eroded = em.clone().erode(neighbours=(5,5,5))
dilatd = em.clone().dilate(neighbours=(5,5,5))

show([(em, __doc__), eroded, dilatd], N=3, viewup='z', zoom=1.4)




"""Density field as a Volume from a point cloud"""

from vedo import *

s  = load(dataurl+'bunny.obj').normalize().subdivide(2).pointSize(2)

vol= s.density().printInfo()

plane = probePlane(vol, normal=(1,1,1)).alpha(0.5)

show([("Point cloud", s),
      ("Point density as Volume", vol, vol.box(), plane) ],
     N=2,
     )


"""Find cells within specified bounds in x, y and/or z."""
from vedo import *

mesh = load(dataurl+'shark.ply').normalize()
mesh.color('aqua').lineWidth(0.1)

z1, z2 = -1.5, -1.0

ids = mesh.findCellsWithin(zbounds=(z1,z2), c='tomato')

#printc('IDs of cells within bounds:\n', ids, c='g')

p1 = Plane(normal=(0,0,1), sx=2, sy=2).z(z1).c('gray',0.5)
p2 = p1.clone().z(z2)

show(mesh, p1, p2, __doc__, axes=9)
